@@ -16,8 +16,6 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <stdio.h>
@@ -26,7 +24,6 @@
 #include <strings.h>
 #include <time.h>
 #include <ctype.h>
-#include <regex.h>
 #include <iostream>
 #include <fstream>
 #include <algorithm>
@@ -44,12 +41,15 @@
 using std::cout;
 using std::cerr;
 using std::endl;
+using std::ios;
+using std::ifstream;
+using std::ofstream;
 using std::string;
+using std::vector;
 using std::set;
 using std::min;
+using std::max;
 
-// This puts a limit to terms length.
-const unsigned int XapianIndex::m_maxTermLength = 240;
 const string XapianIndex::MAGIC_TERM = "X-MetaSE-Doc";
 
 static void removeFirstPosting(Xapian::Document &doc,
@@ -68,7 +68,7 @@ static void removeFirstPosting(Xapian::Document &doc,
 		{
 			// This posting may have been removed already
 #ifdef DEBUG
-			cout << "XapianIndex::removeFirstPosting: " << error.get_errno() << " " << error.get_msg() << endl;
+			cout << "XapianIndex::removeFirstPosting: " << error.get_msg() << endl;
 #endif
 		}
 	}
@@ -113,46 +113,6 @@ XapianIndex &XapianIndex::operator=(const XapianIndex &other)
 	return *this;
 }
 
-string XapianIndex::limitTermLength(const string &term, bool makeUnique)
-{
-	if (term.length() > XapianIndex::m_maxTermLength)
-	{
-		if (makeUnique == false)
-		{
-			// Truncate
-			return term.substr(0, XapianIndex::m_maxTermLength);
-		}
-		else
-		{
-			return StringManip::hashString(term, XapianIndex::m_maxTermLength);
-		}
-	}
-
-	return term;
-}
-
-bool XapianIndex::badField(const string &field)
-{
-	regex_t fieldRegex;
-	regmatch_t pFieldMatches[1];
-	bool isBadField = false;
-
-	// A bad field is one that includes one of our field delimiters
-	if (regcomp(&fieldRegex,
-		"(url|sample|caption|type|timestamp|language)=",
-		REG_EXTENDED|REG_ICASE) == 0)
-	{
-		if (regexec(&fieldRegex, field.c_str(), 1,
-			pFieldMatches, REG_NOTBOL|REG_NOTEOL) == 0)
-		{
-			isBadField = true;
-		}
-	}
-	regfree(&fieldRegex);
-
-	return isBadField;
-}
-
 bool XapianIndex::listDocumentsWithTerm(const string &term, set<unsigned int> &docIds,
 	unsigned int maxDocsCount, unsigned int startDoc) const
 {
@@ -193,7 +153,7 @@ bool XapianIndex::listDocumentsWithTerm(const string &term, set<unsigned int> &d
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't get document list: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't get document list: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -208,27 +168,37 @@ void XapianIndex::addPostingsToDocument(Tokenizer &tokens, Xapian::Document &doc
 	const string &prefix, Xapian::termcount &termPos, StemmingMode mode) const
 {
 	Xapian::Stem *pStemmer = NULL;
+	string stemPrefix("Z");
 	string term;
 
 	// Do we know what language to use for stemming ?
 	if (m_stemLanguage.empty() == false)
 	{
-		pStemmer = new Xapian::Stem(StringManip::toLowerCase(m_stemLanguage));
+		try
+		{
+			pStemmer = new Xapian::Stem(StringManip::toLowerCase(m_stemLanguage));
+		}
+		catch (const Xapian::Error &error)
+		{
+			cerr << "Couldn't create stemmer: " << error.get_type() << ": " << error.get_msg() << endl;
+		}
+	}
+
+	// Stems are Z-prefixed, unless a prefix is already defined
+	if (prefix.empty() == false)
+	{
+		stemPrefix = prefix;
 	}
 
 	// Get the terms
 	while (tokens.nextToken(term) == true)
 	{
+		bool addStem = false;
+
 		if (term.empty() == true)
 		{
 			continue;
 		}
-		// Does it start with a capital letter ?
-		if (isupper((int)term[0]) != 0)
-		{
-			// R-prefix the raw term
-			doc.add_posting(limitTermLength(string("R") + term), termPos);
-		}
 		// Lower case the term
 		term = StringManip::toLowerCase(term);
 
@@ -236,25 +206,30 @@ void XapianIndex::addPostingsToDocument(Tokenizer &tokens, Xapian::Document &doc
 		if ((mode == STORE_UNSTEM) ||
 			(pStemmer == NULL))
 		{
-			doc.add_posting(limitTermLength(prefix + term), termPos);
+			doc.add_posting(prefix + XapianDatabase::limitTermLength(term), termPos);
 		}
 		else if (mode == STORE_STEM)
 		{
-			string stemmedTerm = pStemmer->stem_word(term);
-
-			doc.add_posting(limitTermLength(prefix + stemmedTerm), termPos);
+			addStem = true;
 		}
 		else if (mode == STORE_BOTH)
 		{
-			string stemmedTerm = pStemmer->stem_word(term);
+			// Add both
+			doc.add_posting(prefix + XapianDatabase::limitTermLength(term), termPos);
+			addStem = true;
+		}
 
-			// Add both at the same position
-			doc.add_posting(limitTermLength(prefix + term), termPos);
-			if (stemmedTerm != term)
-			{
-				// No point adding the same term twice
-				doc.add_posting(limitTermLength(prefix + stemmedTerm), termPos);
-			}
+		// Don't stem if the term starts with a digit
+		if ((addStem == true) &&
+			(isdigit((int)term[0]) == 0))
+		{
+#if XAPIAN_MAJOR_VERSION==0
+			string stemmedTerm(pStemmer->stem_word(term));
+#else
+			string stemmedTerm((*pStemmer)(term));
+#endif
+
+			doc.add_term(stemPrefix + XapianDatabase::limitTermLength(stemmedTerm));
 		}
 
 		++termPos;
@@ -274,27 +249,37 @@ void XapianIndex::removeFirstPostingsFromDocument(Tokenizer &tokens, Xapian::Doc
 {
 	Xapian::TermIterator termListIter = doc.termlist_begin();
 	Xapian::Stem *pStemmer = NULL;
+	string stemPrefix("Z");
 	string term;
 
 	// Do we know what language to use for stemming ?
 	if (language.empty() == false)
 	{
-		pStemmer = new Xapian::Stem(StringManip::toLowerCase(language));
+		try
+		{
+			pStemmer = new Xapian::Stem(StringManip::toLowerCase(m_stemLanguage));
+		}
+		catch (const Xapian::Error &error)
+		{
+			cerr << "Couldn't create stemmer: " << error.get_type() << ": " << error.get_msg() << endl;
+		}
+	}
+
+	// Stems are Z-prefixed, unless a prefix is already defined
+	if (prefix.empty() == false)
+	{
+		stemPrefix = prefix;
 	}
 
 	// Get the terms and remove the first posting for each
 	while (tokens.nextToken(term) == true)
 	{
+		bool removeStem = false;
+
 		if (term.empty() == true)
 		{
 			continue;
 		}
-		// Does it start with a capital letter ?
-		if (isupper((int)term[0]) != 0)
-		{
-			// R-prefix the raw term
-			removeFirstPosting(doc, termListIter, string("R") + term);
-		}
 		// Lower case the term
 		term = StringManip::toLowerCase(term);
 
@@ -302,21 +287,34 @@ void XapianIndex::removeFirstPostingsFromDocument(Tokenizer &tokens, Xapian::Doc
 		if ((mode == STORE_UNSTEM) ||
 			(pStemmer == NULL))
 		{
-			removeFirstPosting(doc, termListIter, limitTermLength(prefix + term));
+			removeFirstPosting(doc, termListIter, prefix + XapianDatabase::limitTermLength(term));
 		}
 		else if (mode == STORE_STEM)
 		{
-			removeFirstPosting(doc, termListIter, limitTermLength(prefix + pStemmer->stem_word(term)));
+			removeStem = true;
 		}
 		else if (mode == STORE_BOTH)
 		{
-			string stemmedTerm = pStemmer->stem_word(term);
+			// Remove both
+			removeFirstPosting(doc, termListIter, prefix + XapianDatabase::limitTermLength(term));
+			removeStem = true;
+		}
 
-			removeFirstPosting(doc, termListIter, limitTermLength(prefix + term));
-			if (stemmedTerm != term)
-			{
-				removeFirstPosting(doc, termListIter, limitTermLength(prefix + stemmedTerm));
-			}
+		// Since stems don't have positional information, we can't simply remove them 
+		// since any may appear more than once in the original document
+		// We can only remove those that have some prefix set
+		// Don't stem if the term starts with a digit
+		if ((removeStem == true) &&
+			(prefix.empty() == false) &&
+			(isdigit((int)term[0]) == 0))
+		{
+#if XAPIAN_MAJOR_VERSION==0
+			string stemmedTerm(pStemmer->stem_word(term));
+#else
+			string stemmedTerm((*pStemmer)(term));
+#endif
+
+			doc.remove_term(stemPrefix + XapianDatabase::limitTermLength(stemmedTerm));
 		}
 	}
 
@@ -348,16 +346,23 @@ void XapianIndex::addCommonTerms(const DocumentInfo &info, Xapian::Document &doc
 	}
 
 	// Index the full URL with prefix U
-	doc.add_term(limitTermLength(string("U") + location, true));
+	doc.add_term(string("U") + XapianDatabase::limitTermLength(Url::escapeUrl(location), true));
+	// ...the base file with XFILE:
+	string::size_type qmPos = location.find("?");
+	if ((urlObj.isLocal() == true) &&
+		(qmPos != string::npos))
+	{
+		doc.add_term(string("XFILE:") + XapianDatabase::limitTermLength(Url::escapeUrl(location.substr(0, qmPos)), true));
+	}
 	// ...the host name and included domains with prefix H
 	string hostName(StringManip::toLowerCase(urlObj.getHost()));
 	if (hostName.empty() == false)
 	{
-		doc.add_term(limitTermLength(string("H") + hostName, true));
+		doc.add_term(string("H") + XapianDatabase::limitTermLength(hostName, true));
 		string::size_type dotPos = hostName.find('.');
 		while (dotPos != string::npos)
 		{
-			doc.add_term(limitTermLength(string("H") + hostName.substr(dotPos + 1), true));
+			doc.add_term(string("H") + XapianDatabase::limitTermLength(hostName.substr(dotPos + 1), true));
 
 			// Next
 			dotPos = hostName.find('.', dotPos + 1);
@@ -367,11 +372,15 @@ void XapianIndex::addCommonTerms(const DocumentInfo &info, Xapian::Document &doc
 	string tree(urlObj.getLocation());
 	if (tree.empty() == false)
 	{
-		doc.add_term(limitTermLength(string("XDIR:") + tree, true));
+		doc.add_term(string("XDIR:") + XapianDatabase::limitTermLength(Url::escapeUrl(tree), true));
+		if (tree[0] == '/')
+		{
+			doc.add_term("XDIR:/");
+		}
 		string::size_type slashPos = tree.find('/', 1);
 		while (slashPos != string::npos)
 		{
-			doc.add_term(limitTermLength(string("XDIR:") + tree.substr(0, slashPos), true));
+			doc.add_term(string("XDIR:") + XapianDatabase::limitTermLength(Url::escapeUrl(tree.substr(0, slashPos)), true));
 
 			// Next
 			slashPos = tree.find('/', slashPos + 1);
@@ -383,7 +392,7 @@ void XapianIndex::addCommonTerms(const DocumentInfo &info, Xapian::Document &doc
 	{
 		string extension;
 
-		doc.add_term(limitTermLength(string("P") + fileName, true));
+		doc.add_term(string("P") + XapianDatabase::limitTermLength(Url::escapeUrl(fileName), true));
 
 		// Does it have an extension ?
 		string::size_type extPos = fileName.rfind('.');
@@ -392,12 +401,12 @@ void XapianIndex::addCommonTerms(const DocumentInfo &info, Xapian::Document &doc
 		{
 			extension = StringManip::toLowerCase(fileName.substr(extPos + 1));
 		}
-		doc.add_term(limitTermLength(string("XEXT:") + extension));
+		doc.add_term(string("XEXT:") + XapianDatabase::limitTermLength(extension));
 	}
 	// Add the date terms D, M and Y
 	time_t timeT = TimeConverter::fromTimestamp(info.getTimestamp());
 	struct tm *tm = localtime(&timeT);
-	string yyyymmdd = TimeConverter::toTimestamp(tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
+	string yyyymmdd = TimeConverter::toYYYYMMDDString(tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
 	if (yyyymmdd.length() == 8)
 	{
 		doc.add_term(string("D") + yyyymmdd);
@@ -462,17 +471,25 @@ void XapianIndex::removeCommonTerms(Xapian::Document &doc)
 		removeFirstPostingsFromDocument(titleTokens, doc, "", language, m_stemMode);
 	}
 
-	// Title
-	commonTerms.insert(limitTermLength(string("U") + docInfo.getLocation(), true));
+	// Location 
+	string location(docInfo.getLocation());
+	commonTerms.insert(string("U") + XapianDatabase::limitTermLength(Url::escapeUrl(location), true));
+	// Base file
+	string::size_type qmPos = location.find("?");
+	if ((urlObj.isLocal() == true) &&
+		(qmPos != string::npos))
+	{
+		commonTerms.insert(string("XFILE:") + XapianDatabase::limitTermLength(Url::escapeUrl(location.substr(0, qmPos)), true));
+	}
 	// Host name
 	string hostName(StringManip::toLowerCase(urlObj.getHost()));
 	if (hostName.empty() == false)
 	{
-		commonTerms.insert(limitTermLength(string("H") + hostName, true));
+		commonTerms.insert(string("H") + XapianDatabase::limitTermLength(hostName, true));
 		string::size_type dotPos = hostName.find('.');
 		while (dotPos != string::npos)
 		{
-			commonTerms.insert(limitTermLength(string("H") + hostName.substr(dotPos + 1), true));
+			commonTerms.insert(string("H") + XapianDatabase::limitTermLength(hostName.substr(dotPos + 1), true));
 
 			// Next
 			dotPos = hostName.find('.', dotPos + 1);
@@ -482,11 +499,15 @@ void XapianIndex::removeCommonTerms(Xapian::Document &doc)
 	string tree(urlObj.getLocation());
 	if (tree.empty() == false)
 	{
-		commonTerms.insert(limitTermLength(string("XDIR:") + tree, true));
+		commonTerms.insert(string("XDIR:") + XapianDatabase::limitTermLength(Url::escapeUrl(tree), true));
+		if (tree[0] == '/')
+		{
+			commonTerms.insert("XDIR:/");
+		}
 		string::size_type slashPos = tree.find('/', 1);
 		while (slashPos != string::npos)
 		{
-			commonTerms.insert(limitTermLength(string("XDIR:") + tree.substr(0, slashPos), true));
+			commonTerms.insert(string("XDIR:") + XapianDatabase::limitTermLength(Url::escapeUrl(tree.substr(0, slashPos)), true));
 
 			// Next
 			slashPos = tree.find('/', slashPos + 1);
@@ -498,7 +519,7 @@ void XapianIndex::removeCommonTerms(Xapian::Document &doc)
 	{
 		string extension;
 
-		commonTerms.insert(limitTermLength(string("P") + fileName, true));
+		commonTerms.insert(string("P") + XapianDatabase::limitTermLength(Url::escapeUrl(fileName), true));
 
 		// Does it have an extension ?
 		string::size_type extPos = fileName.rfind('.');
@@ -507,12 +528,12 @@ void XapianIndex::removeCommonTerms(Xapian::Document &doc)
 		{
 			extension = StringManip::toLowerCase(fileName.substr(extPos + 1));
 		}
-		commonTerms.insert(limitTermLength(string("XEXT:") + extension));
+		commonTerms.insert(string("XEXT:") + XapianDatabase::limitTermLength(extension));
 	}
 	// Date terms
 	time_t timeT = TimeConverter::fromTimestamp(docInfo.getTimestamp());
 	struct tm *tm = localtime(&timeT);
-	string yyyymmdd = TimeConverter::toTimestamp(tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
+	string yyyymmdd = TimeConverter::toYYYYMMDDString(tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
 	if (yyyymmdd.length() == 8)
 	{
 		commonTerms.insert(string("D") + yyyymmdd);
@@ -533,7 +554,7 @@ void XapianIndex::removeCommonTerms(Xapian::Document &doc)
 		catch (const Xapian::Error &error)
 		{
 #ifdef DEBUG
-			cout << "XapianIndex::removeCommonTerms: " << error.get_errno() << " " << error.get_msg() << endl;
+			cout << "XapianIndex::removeCommonTerms: " << error.get_msg() << endl;
 #endif
 		}
 	}
@@ -563,7 +584,7 @@ string XapianIndex::scanDocument(const char *pData, unsigned int dataLength,
 		}
 		catch (const Xapian::Error &error)
 		{
-			cerr << "XapianIndex::scanDocument: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+			cerr << "Couldn't create stemmer: " << error.get_type() << ": " << error.get_msg() << endl;
 			continue;
 		}
 
@@ -583,51 +604,58 @@ string XapianIndex::scanDocument(const char *pData, unsigned int dataLength,
 void XapianIndex::setDocumentData(const DocumentInfo &info, Xapian::Document &doc,
 	const string &language) const
 {
-	string title(info.getTitle());
-	string timestamp(info.getTimestamp());
-	char tmpStr[64];
-	time_t timeT = TimeConverter::fromTimestamp(timestamp);
+	time_t timeT = TimeConverter::fromTimestamp(info.getTimestamp());
 
 	// Add this value to allow sorting by date
 	doc.add_value(0, StringManip::integerToBinaryString((uint32_t)timeT));
 
-	// Set the document data omindex-style
-	string record = "url=";
-	record += info.getLocation();
-	// The sample will be generated at query time
-	record += "\nsample=";
-	record += "\ncaption=";
-	if (badField(title) == true)
+	DocumentInfo docCopy(info);
+	docCopy.setLanguage(language);
+	doc.set_data(XapianDatabase::propsToRecord(&docCopy));
+}
+
+bool XapianIndex::deleteDocuments(const string &term)
+{
+	bool unindexed = false;
+
+	if (term.empty() == true)
 	{
-		// Modify the title if necessary
-		string::size_type pos = title.find("=");
-		while (pos != string::npos)
-		{
-			title[pos] = ' ';
-			pos = title.find("=", pos + 1);
-		}
-#ifdef DEBUG
-		cout << "XapianIndex::setDocumentData: modified title" << endl;
-#endif
+		return false;
 	}
-	record += title;
-	record += "\ntype=";
-	record += info.getType();
-	// Append a timestamp, in a format compatible with Omega
-	record += "\nmodtime=";
-	snprintf(tmpStr, 64, "%ld", timeT);
-	record += tmpStr;
-	// ...and the language
-	record += "\nlanguage=";
-	record += StringManip::toLowerCase(language);
-	// ...and the file size
-	record += "\nsize=";
-	snprintf(tmpStr, 64, "%ld", info.getSize());
-	record += tmpStr;
+
+	XapianDatabase *pDatabase = XapianDatabaseFactory::getDatabase(m_databaseName, false);
+	if (pDatabase == NULL)
+	{
+		cerr << "Bad index " << m_databaseName << endl;
+		return false;
+	}
+
+	try
+	{
+		Xapian::WritableDatabase *pIndex = pDatabase->writeLock();
+		if (pIndex != NULL)
+		{
 #ifdef DEBUG
-	cout << "XapianIndex::setDocumentData: document data is " << record << endl;
+			cout << "XapianIndex::deleteDocuments: term is " << term << endl;
 #endif
-	doc.set_data(record);
+
+			// Delete documents from the index
+			pIndex->delete_document(term);
+
+			unindexed = true;
+		}
+	}
+	catch (const Xapian::Error &error)
+	{
+		cerr << "Couldn't unindex documents: " << error.get_type() << ": " << error.get_msg() << endl;
+	}
+	catch (...)
+	{
+		cerr << "Couldn't unindex documents, unknown exception occured" << endl;
+	}
+	pDatabase->unlock();
+
+	return unindexed;
 }
 
 //
@@ -640,6 +668,43 @@ bool XapianIndex::isGood(void) const
 	return m_goodIndex;
 }
 
+/// Gets the version number.
+double XapianIndex::getVersion(void) const
+{
+	ifstream verFile;
+	string verFileName(m_databaseName);
+	double version = 0;
+
+	verFileName += "/version";
+	verFile.open(verFileName.c_str());
+	if (verFile.good() == true)
+	{
+		verFile >> version;
+	}
+	verFile.close();
+
+	return version;
+}
+
+/// Sets the version number.
+bool XapianIndex::setVersion(double version) const
+{
+	ofstream verFile;
+	string verFileName(m_databaseName);
+	bool setVer = false;
+
+	verFileName += "/version";
+	verFile.open(verFileName.c_str(), ios::trunc);
+	if (verFile.good() == true)
+	{
+		verFile << version;
+		setVer = true;
+	}
+	verFile.close();
+
+	return setVer;
+}
+
 /// Gets the index location.
 string XapianIndex::getLocation(void) const
 {
@@ -704,7 +769,7 @@ bool XapianIndex::getDocumentInfo(unsigned int docId, DocumentInfo &docInfo) con
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't get document properties: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't get document properties: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -715,6 +780,44 @@ bool XapianIndex::getDocumentInfo(unsigned int docId, DocumentInfo &docInfo) con
 	return foundDocument;
 }
 
+/// Returns a document's terms count.
+unsigned int XapianIndex::getDocumentTermsCount(unsigned int docId) const
+{
+	unsigned int termsCount = 0;
+
+	XapianDatabase *pDatabase = XapianDatabaseFactory::getDatabase(m_databaseName);
+	if (pDatabase == NULL)
+	{
+		cerr << "Bad index " << m_databaseName << endl;
+		return 0;
+	}
+
+	try
+	{
+		Xapian::Database *pIndex = pDatabase->readLock();
+		if (pIndex != NULL)
+		{
+			Xapian::Document doc = pIndex->get_document(docId);
+
+			termsCount = doc.termlist_count();
+#ifdef DEBUG
+			cout << "XapianIndex::getDocumentTermsCount: " << termsCount << " terms in document " << docId << endl;
+#endif
+		}
+	}
+	catch (const Xapian::Error &error)
+	{
+		cerr << "Couldn't get document terms count: " << error.get_type() << ": " << error.get_msg() << endl;
+	}
+	catch (...)
+	{
+		cerr << "Couldn't get document terms count, unknown exception occured" << endl;
+	}
+	pDatabase->unlock();
+
+	return termsCount;
+}
+
 /// Determines whether a document has a label.
 bool XapianIndex::hasLabel(unsigned int docId, const string &name) const
 {
@@ -736,7 +839,7 @@ bool XapianIndex::hasLabel(unsigned int docId, const string &name) const
 
 			// Get documents that have this label
 			// FIXME: would it be faster to get the document's terms ?
-			term += name;
+			term += XapianDatabase::limitTermLength(Url::escapeUrl(name));
 			Xapian::PostingIterator postingIter = pIndex->postlist_begin(term);
 			if (postingIter != pIndex->postlist_end(term))
 			{
@@ -752,7 +855,7 @@ bool XapianIndex::hasLabel(unsigned int docId, const string &name) const
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't check document labels: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't check document labels: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -795,7 +898,7 @@ bool XapianIndex::getDocumentLabels(unsigned int docId, set<string> &labels) con
 					// Is this a label ?
 					if (strncasecmp((*termIter).c_str(), "XLABEL:", min(7, (int)(*termIter).length())) == 0)
 					{
-						labels.insert((*termIter).substr(7));
+						labels.insert(Url::unescapeUrl((*termIter).substr(7)));
 					}
 				}
 				gotLabels = true;
@@ -804,7 +907,7 @@ bool XapianIndex::getDocumentLabels(unsigned int docId, set<string> &labels) con
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't get document's labels: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't get document's labels: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -832,7 +935,7 @@ unsigned int XapianIndex::hasDocument(const string &url) const
 		Xapian::Database *pIndex = pDatabase->readLock();
 		if (pIndex != NULL)
 		{
-			string term = limitTermLength(string("U") + Url::canonicalizeUrl(url), true);
+			string term = string("U") + XapianDatabase::limitTermLength(Url::escapeUrl(Url::canonicalizeUrl(url)), true);
 
 			// Get documents that have this term
 			Xapian::PostingIterator postingIter = pIndex->postlist_begin(term);
@@ -850,7 +953,7 @@ unsigned int XapianIndex::hasDocument(const string &url) const
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't look for document: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't look for document: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -883,14 +986,6 @@ unsigned int XapianIndex::getCloseTerms(const string &term, set<string> &suggest
 			{
 				string baseTerm(StringManip::toLowerCase(term));
 				unsigned int count = 0;
-				bool isUpper = false;
-
-				if (isupper((int)term[0]) != 0)
-				{
-					// R-prefix the term
-					baseTerm = string("R") + term;
-					isUpper = true;
-				}
 
 				// Get the next 10 terms
 				for (termIter.skip_to(baseTerm);
@@ -901,28 +996,9 @@ unsigned int XapianIndex::getCloseTerms(const string &term, set<string> &suggest
 					if (suggestedTerm.find(baseTerm) != 0)
 					{
 						// This term doesn't have the same root
-						if (isUpper == true)
-						{
-							// Try again without capital letters
-							baseTerm = StringManip::toLowerCase(term);
-							termIter = pIndex->allterms_begin();
-							if (termIter != pIndex->allterms_end())
-							{
-								termIter.skip_to(baseTerm);
-								isUpper = false;
-								continue;
-							}
-						}
-
 						break;
 					}
 
-					if (isUpper == true)
-					{
-						// Remove the R prefix
-						suggestedTerm.erase(0, 1);
-					}
-
 					suggestions.insert(suggestedTerm);
 					++count;
 				}
@@ -931,7 +1007,7 @@ unsigned int XapianIndex::getCloseTerms(const string &term, set<string> &suggest
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't get terms: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't get terms: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -964,7 +1040,7 @@ unsigned int XapianIndex::getLastDocumentID(void) const
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't get last document ID: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't get last document ID: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -1002,13 +1078,14 @@ unsigned int XapianIndex::getDocumentsCount(const string &labelName) const
 
 				// Each label appears only one per document so the collection frequency
 				// is the number of documents that have this label
-				docCount = pIndex->get_collection_freq(term + labelName);
+				term += XapianDatabase::limitTermLength(Url::escapeUrl(labelName));
+				docCount = pIndex->get_collection_freq(term);
 			}
 		}
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't count documents: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't count documents: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -1023,7 +1100,7 @@ unsigned int XapianIndex::getDocumentsCount(const string &labelName) const
 unsigned int XapianIndex::listDocuments(set<unsigned int> &docIds,
 	unsigned int maxDocsCount, unsigned int startDoc) const
 {
-	// Get documents that have the magic term
+	// All documents have the magic term
 	return listDocumentsWithTerm(MAGIC_TERM, docIds, maxDocsCount, startDoc);
 }
 
@@ -1033,7 +1110,7 @@ bool XapianIndex::listDocumentsWithLabel(const string &name, set<unsigned int> &
 {
 	string term("XLABEL:");
 
-	term += name;
+	term += XapianDatabase::limitTermLength(Url::escapeUrl(name));
 	return listDocumentsWithTerm(term, docIds, maxDocsCount, startDoc);
 }
 
@@ -1041,10 +1118,10 @@ bool XapianIndex::listDocumentsWithLabel(const string &name, set<unsigned int> &
 bool XapianIndex::listDocumentsInDirectory(const string &dirName, set<unsigned int> &docIds,
 	unsigned int maxDocsCount, unsigned int startDoc) const
 {
-	string term("XDIR:");
+	string term(Url::escapeUrl(dirName));
 
-	term += dirName;
-	return listDocumentsWithTerm(term, docIds, maxDocsCount, startDoc);
+	return listDocumentsWithTerm(string("XDIR") + XapianDatabase::limitTermLength(term, true),
+		docIds, maxDocsCount, startDoc);
 }
 
 /// Indexes the given data.
@@ -1096,7 +1173,7 @@ bool XapianIndex::indexDocument(Tokenizer &tokens, const std::set<std::string> &
 		for (set<string>::const_iterator labelIter = labels.begin(); labelIter != labels.end();
 			++labelIter)
 		{
-			doc.add_term(limitTermLength(string("XLABEL:") + *labelIter));
+			doc.add_term(string("XLABEL:") + XapianDatabase::limitTermLength(Url::escapeUrl(*labelIter)));
 		}
 		addCommonTerms(docInfo, doc, termPos);
 		setDocumentData(docInfo, doc, m_stemLanguage);
@@ -1111,7 +1188,7 @@ bool XapianIndex::indexDocument(Tokenizer &tokens, const std::set<std::string> &
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't index document: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't index document: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -1176,7 +1253,7 @@ bool XapianIndex::updateDocument(unsigned int docId, Tokenizer &tokens)
 			for (set<string>::const_iterator labelIter = labels.begin(); labelIter != labels.end();
 				++labelIter)
 			{
-				doc.add_term(limitTermLength(string("XLABEL:") + *labelIter));
+				doc.add_term(string("XLABEL:") + XapianDatabase::limitTermLength(Url::escapeUrl(*labelIter)));
 			}
 		}
 		addCommonTerms(docInfo, doc, termPos);
@@ -1192,7 +1269,7 @@ bool XapianIndex::updateDocument(unsigned int docId, Tokenizer &tokens)
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't update document: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't update document: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -1240,7 +1317,7 @@ bool XapianIndex::updateDocumentInfo(unsigned int docId, const DocumentInfo &doc
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't update document properties: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't update document properties: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -1312,7 +1389,7 @@ bool XapianIndex::setDocumentsLabels(const set<unsigned int> &docIds,
 			{
 				if (labelIter->empty() == false)
 				{
-					doc.add_term(limitTermLength(string("XLABEL:") + *labelIter));
+					doc.add_term(string("XLABEL:") + XapianDatabase::limitTermLength(Url::escapeUrl(*labelIter)));
 				}
 			}
 
@@ -1321,7 +1398,7 @@ bool XapianIndex::setDocumentsLabels(const set<unsigned int> &docIds,
 		}
 		catch (const Xapian::Error &error)
 		{
-			cerr << "Couldn't update document's labels: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+			cerr << "Couldn't update document's labels: " << error.get_type() << ": " << error.get_msg() << endl;
 		}
 		catch (...)
 		{
@@ -1363,7 +1440,7 @@ bool XapianIndex::unindexDocument(unsigned int docId)
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't unindex document: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't unindex document: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -1377,86 +1454,37 @@ bool XapianIndex::unindexDocument(unsigned int docId)
 /// Unindexes the given document.
 bool XapianIndex::unindexDocument(const string &location)
 {
-	bool unindexed = false;
+	string term(string("U") + XapianDatabase::limitTermLength(Url::escapeUrl(Url::canonicalizeUrl(location)), true));
 
-	if (location.empty() == true)
-	{
-		return false;
-	}
-
-	XapianDatabase *pDatabase = XapianDatabaseFactory::getDatabase(m_databaseName, false);
-	if (pDatabase == NULL)
-	{
-		cerr << "Bad index " << m_databaseName << endl;
-		return false;
-	}
-
-	try
-	{
-		Xapian::WritableDatabase *pIndex = pDatabase->writeLock();
-		if (pIndex != NULL)
-		{
-			string term = limitTermLength(string("U") + Url::canonicalizeUrl(location), true);
-
-			// Only one document should have this term
-			pIndex->delete_document(term);
-			unindexed = true;
-		}
-	}
-	catch (const Xapian::Error &error)
-	{
-		cerr << "Couldn't unindex documents: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
-	}
-	catch (...)
-	{
-		cerr << "Couldn't unindex documents, unknown exception occured" << endl;
-	}
-	pDatabase->unlock();
-
-	return unindexed;
+	return deleteDocuments(term);
 }
 
-/// Unindexes documents with the given label.
-bool XapianIndex::unindexDocuments(const string &labelName)
+/// Unindexes documents.
+bool XapianIndex::unindexDocuments(const string &name, NameType type)
 {
-	bool unindexed = false;
+	string term;
 
-	if (labelName.empty() == true)
+	if (type == BY_LABEL)
 	{
-		return false;
+		term = string("XLABEL:") + XapianDatabase::limitTermLength(Url::escapeUrl(name));
+	}
+	else if (type == BY_DIRECTORY)
+	{
+		term = string("XDIR:") + XapianDatabase::limitTermLength(Url::escapeUrl(name), true);
+	}
+	else if (type == BY_FILE)
+	{
+		term = string("XFILE:") + XapianDatabase::limitTermLength(Url::escapeUrl(name), true);
 	}
 
-	XapianDatabase *pDatabase = XapianDatabaseFactory::getDatabase(m_databaseName, false);
-	if (pDatabase == NULL)
-	{
-		cerr << "Bad index " << m_databaseName << endl;
-		return false;
-	}
+	return deleteDocuments(term);
+}
 
-	try
-	{
-		Xapian::WritableDatabase *pIndex = pDatabase->writeLock();
-		if (pIndex != NULL)
-		{
-			string term("XLABEL:");
-
-			// Delete documents from the index
-			term += labelName;
-			pIndex->delete_document(term);
-			unindexed = true;
-		}
-	}
-	catch (const Xapian::Error &error)
-	{
-		cerr << "Couldn't unindex documents: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
-	}
-	catch (...)
-	{
-		cerr << "Couldn't unindex documents, unknown exception occured" << endl;
-	}
-	pDatabase->unlock();
-
-	return unindexed;
+/// Unindexes all documents.
+bool XapianIndex::unindexAllDocuments(void)
+{
+	// All documents have the magic term
+	return deleteDocuments(MAGIC_TERM);
 }
 
 /// Renames a label.
@@ -1479,7 +1507,7 @@ bool XapianIndex::renameLabel(const string &name, const string &newName)
 			string term("XLABEL:");
 
 			// Get documents that have this label
-			term += name;
+			term += XapianDatabase::limitTermLength(Url::escapeUrl(name));
 			for (Xapian::PostingIterator postingIter = pIndex->postlist_begin(term);
 				postingIter != pIndex->postlist_end(term); ++postingIter)
 			{
@@ -1490,7 +1518,7 @@ bool XapianIndex::renameLabel(const string &name, const string &newName)
 				// Remove the term
 				doc.remove_term(term);
 				// ...add the new one
-				doc.add_term(limitTermLength(string("XLABEL:") + newName));
+				doc.add_term(string("XLABEL:") + XapianDatabase::limitTermLength(Url::escapeUrl(newName)));
 				// ...and update the document
 				pIndex->replace_document(docId, doc);
 			}
@@ -1500,7 +1528,7 @@ bool XapianIndex::renameLabel(const string &name, const string &newName)
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't delete label: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't delete label: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -1531,7 +1559,7 @@ bool XapianIndex::deleteLabel(const string &name)
 			string term("XLABEL:");
 
 			// Get documents that have this label
-			term += name;
+			term += XapianDatabase::limitTermLength(Url::escapeUrl(name));
 			for (Xapian::PostingIterator postingIter = pIndex->postlist_begin(term);
 				postingIter != pIndex->postlist_end(term); ++postingIter)
 			{
@@ -1549,7 +1577,7 @@ bool XapianIndex::deleteLabel(const string &name)
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't delete label: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't delete label: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
@@ -1583,7 +1611,7 @@ bool XapianIndex::flush(void)
 	}
 	catch (const Xapian::Error &error)
 	{
-		cerr << "Couldn't flush database: " << error.get_type() << ": " << error.get_errno() << " " << error.get_msg() << endl;
+		cerr << "Couldn't flush database: " << error.get_type() << ": " << error.get_msg() << endl;
 	}
 	catch (...)
 	{
