@@ -17,6 +17,8 @@
  */
 
 #include <stdlib.h>
+#include <map>
+#include <set>
 #include <iostream>
 #include <boost/spirit/core.hpp>
 #include <boost/spirit/actor/push_back_actor.hpp>
@@ -25,15 +27,21 @@
 
 #include "StringManip.h"
 #include "Url.h"
-#include "HtmlTokenizer.h"
-#include "XmlTokenizer.h"
+#include "HtmlFilter.h"
+#include "FilterFactory.h"
 #include "FileCollector.h"
+#include "FilterUtils.h"
 #include "SherlockParser.h"
 
 using std::cout;
 using std::endl;
 using std::string;
+using std::vector;
 using std::map;
+using std::set;
+using std::exception;
+
+using namespace boost::spirit;
 
 // A function object to lower case map keys with for_each()
 struct LowerAndCopy
@@ -53,15 +61,12 @@ struct LowerAndCopy
 
 };
 
-using namespace std;
-using namespace boost::spirit;
-
-struct skip_grammar : public grammar<skip_grammar>
+struct plugin_skip_grammar : public grammar<plugin_skip_grammar>
 {
 	template <typename ScannerT>
 	struct definition
 	{
-		definition(skip_grammar const &self)
+		definition(plugin_skip_grammar const &self)
 		{
 			// Skip all spaces and comments, starting with a #
 			// FIXME: make sure comments start at the beginning of the line !
@@ -139,7 +144,7 @@ struct plugin_min_grammar : public grammar<plugin_min_grammar>
   */
 struct plugin_grammar : public grammar<plugin_grammar>
 {
-	 plugin_grammar(map<string, string> &searchParams,
+	plugin_grammar(map<string, string> &searchParams,
 		map<string, string> &interpretParams,
 		map<string, string> &inputItems,
 		string &userInput, string &nextInput,
@@ -268,7 +273,7 @@ SherlockResponseParser::~SherlockResponseParser()
 {
 }
 
-bool SherlockResponseParser::parse(const Document *pResponseDoc, vector<Result> &resultsList,
+bool SherlockResponseParser::parse(const Document *pResponseDoc, vector<DocumentInfo> &resultsList,
 	unsigned int &totalResults, unsigned int &firstResultIndex) const
 {
 	float pseudoScore = 100;
@@ -360,40 +365,51 @@ bool SherlockResponseParser::parse(const Document *pResponseDoc, vector<Result>
 #endif
 			chunkDoc.setData(dummyHtml.c_str(), dummyHtml.length());
 
-			HtmlTokenizer chunkTokens(&chunkDoc, false, true);
-			set<Link> &chunkLinks = chunkTokens.getLinks();
-			unsigned int endOfFirstLink = 0, startOfSecondLink = 0, endOfSecondLink = 0, startOfThirdLink = 0;
-
-			// The result's URL and title should be given by the first link
-			for (set<Link>::iterator linkIter = chunkLinks.begin(); linkIter != chunkLinks.end(); ++linkIter)
+			// Feed this chunk to the filter
+			Dijon::HtmlFilter chunkFilter("text/html");
+			set<Dijon::Link> chunkLinks;
+			if ((FilterUtils::feedFilter(chunkDoc, &chunkFilter) == true) &&
+				(chunkFilter.get_links(chunkLinks) == true) &&
+				(chunkFilter.next_document() == true))
 			{
-				if (linkIter->m_index == 0)
+				unsigned int endOfFirstLink = 0, startOfSecondLink = 0, endOfSecondLink = 0, startOfThirdLink = 0;
+
+				// The result's URL and title should be given by the first link
+				for (set<Dijon::Link>::iterator linkIter = chunkLinks.begin(); linkIter != chunkLinks.end(); ++linkIter)
 				{
-					url = linkIter->m_url;
-					name = linkIter->m_name;
+					if (linkIter->m_index == 0)
+					{
+						url = linkIter->m_url;
+						name = linkIter->m_name;
 #ifdef DEBUG
-					cout << "SherlockResponseParser::parse: first link in chunk is "
-						<< url << endl;
+						cout << "SherlockResponseParser::parse: first link in chunk is "
+							<< url << endl;
 #endif
-					endOfFirstLink = linkIter->m_endPos;
+						endOfFirstLink = linkIter->m_endPos;
+					}
+					else if (linkIter->m_index == 1)
+					{
+						startOfSecondLink = linkIter->m_startPos;
+						endOfSecondLink = linkIter->m_endPos;
+					}
+					else if (linkIter->m_index == 2)
+					{
+						startOfThirdLink = linkIter->m_startPos;
+					}
 				}
-				else if (linkIter->m_index == 1)
-				{
-					startOfSecondLink = linkIter->m_startPos;
-					endOfSecondLink = linkIter->m_endPos;
-				}
-				else if (linkIter->m_index == 2)
-				{
-					startOfThirdLink = linkIter->m_startPos;
-				}
-			}
 
-			// Any extract ?
-			extract = chunkTokens.getAbstract();
-			if (extract.empty() == true)
-			{
-				extract = XmlTokenizer::stripTags(resultItem);
-				StringManip::trimSpaces(extract);
+				// Any extract ?
+				const map<string, string> &metaData = chunkFilter.get_meta_data();
+				map<string, string>::const_iterator abstractIter = metaData.find("abstract");
+				if (abstractIter == metaData.end())
+				{
+					extract = FilterUtils::stripMarkup(resultItem);
+					StringManip::trimSpaces(extract);
+				}
+				else
+				{
+					extract = abstractIter->second;
+				}
 			}
 		}
 		else
@@ -444,7 +460,11 @@ bool SherlockResponseParser::parse(const Document *pResponseDoc, vector<Result>
 				url = tmpUrl;
 			}
 
-			resultsList.push_back(Result(url, name, extract, "", pseudoScore));
+			DocumentInfo result(name, url, "", "");
+			result.setExtract(extract);
+			result.setScore(pseudoScore);
+
+			resultsList.push_back(result);
 			--pseudoScore;
 			foundResult = true;
 			if (resultsList.size() == totalResults)
@@ -510,7 +530,7 @@ ResponseParserInterface *SherlockParser::parse(SearchPluginProperties &propertie
 		{
 			if (extractSearchParams == false)
 			{
-				skip_grammar skip;
+				plugin_skip_grammar skip;
 				plugin_grammar plugin(searchParams, interpretParams, inputItems,
 					userInput, nextInput, nextFactor, nextValue);
 
@@ -519,7 +539,7 @@ ResponseParserInterface *SherlockParser::parse(SearchPluginProperties &propertie
 			}
 			else
 			{
-				skip_grammar skip;
+				plugin_skip_grammar skip;
 				plugin_min_grammar plugin(searchParams);
 
 				parse_info<> parseInfo = boost::spirit::parse(pData, plugin, skip);
