@@ -28,11 +28,13 @@
 #include <iostream>
 
 #include <glibmm/convert.h>
+#include <glibmm/date.h>
 #include <libxml++/parsers/domparser.h>
 #include <libxml++/nodes/node.h>
 #include <libxml++/nodes/textnode.h>
 
 #include "config.h"
+#include "CommandLine.h"
 #include "Languages.h"
 #include "NLS.h"
 #include "StringManip.h"
@@ -85,7 +87,7 @@ static Element *addChildElement(Element *pElem, const string &nodeName, const st
 }
 
 PinotSettings PinotSettings::m_instance;
-bool PinotSettings::m_enableDBus = false;
+bool PinotSettings::m_clientMode = false;
 
 PinotSettings::PinotSettings() :
 	m_xPos(0),
@@ -100,11 +102,15 @@ PinotSettings::PinotSettings() :
 	m_newResultsColourRed(65535),
 	m_newResultsColourGreen(0),
 	m_newResultsColourBlue(0),
+	m_proxyPort(8080),
+	m_proxyEnabled(false),
+	m_isBlackList(true),
 	m_firstRun(false)
 {
-	// Find out if there is a .pinot directory
+	string directoryName(getConfigurationDirectory());
 	struct stat fileStat;
-	string directoryName = getConfigurationDirectory();
+
+	// Find out if there is a .pinot directory
 	if (stat(directoryName.c_str(), &fileStat) != 0)
 	{
 		// No, create it then
@@ -124,10 +130,6 @@ PinotSettings::PinotSettings() :
 	m_docsIndexLocation += "/index";
 	m_daemonIndexLocation = directoryName;
 	m_daemonIndexLocation += "/daemon";
-
-	// The location of the history database is not configurable
-	m_historyDatabase = directoryName;
-	m_historyDatabase += "/history";
 }
 
 PinotSettings::~PinotSettings()
@@ -139,11 +141,11 @@ PinotSettings &PinotSettings::getInstance(void)
 	return m_instance;
 }
 
-bool PinotSettings::enableDBus(bool enable)
+bool PinotSettings::enableClientMode(bool enable)
 {
-	bool isEnabled = m_enableDBus;
+	bool isEnabled = m_clientMode;
 
-	m_enableDBus = enable;
+	m_clientMode = enable;
 
 	return isEnabled;
 }
@@ -196,6 +198,47 @@ string PinotSettings::getCurrentUserName(void)
 	return "";
 }
 
+void PinotSettings::checkHistoryDatabase(void)
+{
+	string uiHistoryDatabase(getConfigurationDirectory());
+	string daemonHistoryDatabase(getConfigurationDirectory());
+	struct stat fileStat;
+
+	// We only need to copy the UI's over to the daemon's history if it doesn't exist
+	uiHistoryDatabase += "/history";
+	daemonHistoryDatabase += "/history-daemon";
+	if ((stat(daemonHistoryDatabase.c_str(), &fileStat) != 0) ||
+		(!S_ISREG(fileStat.st_mode)))
+	{
+		string output;
+
+		CommandLine::runSync(string("\\cp -f ") + uiHistoryDatabase + " " + daemonHistoryDatabase, output);
+#ifdef DEBUG
+		cout << "PinotSettings::checkHistoryDatabase: " << output << endl;
+#endif
+	}
+}
+
+string PinotSettings::getHistoryDatabaseName(bool needToQueryDaemonHistory)
+{
+	string historyDatabase(getConfigurationDirectory());
+
+	if ((m_clientMode == false) ||
+		(needToQueryDaemonHistory == true))
+	{
+		historyDatabase += "/history-daemon";
+	}
+	else
+	{
+		historyDatabase += "/history";
+	}
+#ifdef DEBUG
+	cout << "PinotSettings::getHistoryDatabaseName: " << historyDatabase << endl;
+#endif
+
+	return historyDatabase;
+}
+
 bool PinotSettings::isFirstRun(void) const
 {
 	return m_firstRun;
@@ -210,9 +253,8 @@ void PinotSettings::clear(void)
 	m_engineIds.clear();
 	m_queries.clear();
 	m_labels.clear();
-	m_mailAccounts.clear();
 	m_indexableLocations.clear();
-	m_filePatternsBlackList.clear();
+	m_filePatternsList.clear();
 	m_cacheProviders.clear();
 }
 
@@ -247,26 +289,25 @@ bool PinotSettings::load(void)
 		m_labels.insert(_("New"));
 		m_labels.insert(_("Personal"));
 		// Skip common image, video and archive types
-		m_filePatternsBlackList.insert("*.Z");
-		m_filePatternsBlackList.insert("*.avi");
-		m_filePatternsBlackList.insert("*.asf");
-		m_filePatternsBlackList.insert("*.bz2");
-		m_filePatternsBlackList.insert("*.deb");
-		m_filePatternsBlackList.insert("*.gif");
-		m_filePatternsBlackList.insert("*.gz");
-		m_filePatternsBlackList.insert("*.jpeg");
-		m_filePatternsBlackList.insert("*.jpg");
-		m_filePatternsBlackList.insert("*.png");
-		m_filePatternsBlackList.insert("*.lha");
-		m_filePatternsBlackList.insert("*.mov");
-		m_filePatternsBlackList.insert("*.msf");
-		m_filePatternsBlackList.insert("*.mpeg");
-		m_filePatternsBlackList.insert("*.mpg");
-		m_filePatternsBlackList.insert("*.rpm");
-		m_filePatternsBlackList.insert("*.sh");
-		m_filePatternsBlackList.insert("*.tar");
-		m_filePatternsBlackList.insert("*.wmv");
-		m_filePatternsBlackList.insert("*.zip");
+		m_filePatternsList.insert("*.Z");
+		m_filePatternsList.insert("*.avi");
+		m_filePatternsList.insert("*.asf");
+		m_filePatternsList.insert("*.gif");
+		m_filePatternsList.insert("*.iso");
+		m_filePatternsList.insert("*.jpeg");
+		m_filePatternsList.insert("*.jpg");
+		m_filePatternsList.insert("*.lha");
+		m_filePatternsList.insert("*.mov");
+		m_filePatternsList.insert("*.msf");
+		m_filePatternsList.insert("*.mpeg");
+		m_filePatternsList.insert("*.mpg");
+		m_filePatternsList.insert("*.png");
+		m_filePatternsList.insert("*.rar");
+		m_filePatternsList.insert("*.sh");
+		m_filePatternsList.insert("*.tiff");
+		m_filePatternsList.insert("*.wmv");
+		m_filePatternsList.insert("*.xbm");
+		m_filePatternsList.insert("*.xpm");
 	}
 
 	// Some search engines are hardcoded
@@ -401,15 +442,16 @@ bool PinotSettings::loadConfiguration(const std::string &fileName, bool isGlobal
 				{
 					loadColour(pElem);
 				}
-				else if (nodeName == "mailaccount")
+				else if (nodeName == "proxy")
 				{
-					loadMailAccounts(pElem);
+					loadProxy(pElem);
 				}
 				else if (nodeName == "indexable")
 				{
 					loadIndexableLocations(pElem);
 				}
-				else if (nodeName == "blacklist")
+				else if ((nodeName == "blacklist") ||
+					(nodeName == "patterns"))
 				{
 					loadFilePatterns(pElem);
 				}
@@ -594,7 +636,12 @@ bool PinotSettings::loadQueries(const Element *pElem)
 	}
 
 	QueryProperties queryProps;
+	Date minDate, maxDate;
 	string backCompatString;
+	bool enableMinDate = false, enableMaxDate = false;
+
+	minDate.set_time(time(NULL));
+	maxDate.set_time(time(NULL));
 
 	// Load the query's properties
 	for (Node::NodeList::iterator iter = childNodes.begin(); iter != childNodes.end(); ++iter)
@@ -702,6 +749,32 @@ bool PinotSettings::loadQueries(const Element *pElem)
 			int count = atoi(nodeContent.c_str());
 			queryProps.setMaximumResultsCount((unsigned int)max(count, 10));
 		}
+		else if (nodeName == "enablemindate")
+		{
+			if (nodeContent == "YES")
+			{
+				enableMinDate = true;
+			}
+			queryProps.setMinimumDate(enableMinDate, minDate.get_day(), minDate.get_month(), minDate.get_year());
+		}
+		else if (nodeName == "mindate")
+		{
+			minDate.set_parse(nodeContent);
+			queryProps.setMinimumDate(enableMinDate, minDate.get_day(), minDate.get_month(), minDate.get_year());
+		}
+		else if (nodeName == "enablemaxdate")
+		{
+			if (nodeContent == "YES")
+			{
+				enableMaxDate = true;
+			}
+			queryProps.setMaximumDate(enableMaxDate, maxDate.get_day(), maxDate.get_month(), maxDate.get_year());
+		}
+		else if (nodeName == "maxdate")
+		{
+			maxDate.set_parse(nodeContent);
+			queryProps.setMaximumDate(enableMaxDate, maxDate.get_day(), maxDate.get_month(), maxDate.get_year());
+		}
 		else if (nodeName == "index")
 		{
 			if (nodeContent == "ALL")
@@ -812,7 +885,7 @@ bool PinotSettings::loadColour(const Element *pElem)
 	return true;
 }
 
-bool PinotSettings::loadMailAccounts(const Element *pElem)
+bool PinotSettings::loadProxy(const Element *pElem)
 {
 	if (pElem == NULL)
 	{
@@ -825,9 +898,6 @@ bool PinotSettings::loadMailAccounts(const Element *pElem)
 		return false;
 	}
 
-	TimestampedItem mailAccount;
-
-	// Load the mail account's properties
 	for (Node::NodeList::iterator iter = childNodes.begin(); iter != childNodes.end(); ++iter)
 	{
 		Node *pNode = (*iter);
@@ -840,19 +910,29 @@ bool PinotSettings::loadMailAccounts(const Element *pElem)
 		string nodeName = pElem->get_name();
 		string nodeContent = getElementContent(pElem);
 
-		if (nodeName == "name")
+		if (nodeName == "address")
 		{
-			mailAccount.m_name = nodeContent;
+			m_proxyAddress = nodeContent;
 		}
-		else if (nodeName == "mtime")
+		else if (nodeName == "port")
 		{
-			mailAccount.m_modTime = (time_t)atoi(nodeContent.c_str());
+			m_proxyPort = (unsigned int)atoi(nodeContent.c_str());
 		}
-	}
-
-	if (mailAccount.m_name.empty() == false)
-	{
-		m_mailAccounts.insert(mailAccount);
+		else if (nodeName == "type")
+		{
+			m_proxyType = nodeContent;
+		}
+		else if (nodeName == "enable")
+		{
+			if (nodeContent == "YES")
+			{
+				m_proxyEnabled = true;
+			}
+			else
+			{
+				m_proxyEnabled = false;
+			}
+		}	
 	}
 
 	return true;
@@ -939,7 +1019,18 @@ bool PinotSettings::loadFilePatterns(const Element *pElem)
 
 		if (nodeName == "pattern")
 		{
-			m_filePatternsBlackList.insert(nodeContent);
+			m_filePatternsList.insert(nodeContent);
+		}
+		else if (nodeName == "forbid")
+		{
+			if (nodeContent == "YES")
+			{
+				m_isBlackList = true;
+			}
+			else
+			{
+				m_isBlackList = false;
+			}
 		}
 	}
 
@@ -1145,10 +1236,28 @@ bool PinotSettings::save(void)
 		{
 			return false;
 		}
+
+		Date aDate;
+		unsigned int day, month, year;
+
 		addChildElement(pElem, "name", queryIter->first);
 		addChildElement(pElem, "text", queryIter->second.getFreeQuery());
 		sprintf(numStr, "%u", queryIter->second.getMaximumResultsCount());
 		addChildElement(pElem, "maxresults", numStr);
+		bool enable = queryIter->second.getMinimumDate(day, month, year);
+		if (year > 0)
+		{
+			addChildElement(pElem, "enablemindate", (enable ? "YES" : "NO"));
+			aDate.set_dmy((Date::Day )day, (Date::Month )month, (Date::Year )year);
+			addChildElement(pElem, "mindate", aDate.format_string("%F"));
+		}
+		enable = queryIter->second.getMaximumDate(day, month, year);
+		if (year > 0)
+		{
+			addChildElement(pElem, "enablemaxdate", (enable ? "YES" : "NO"));
+			aDate.set_dmy((Date::Day )day, (Date::Month )month, (Date::Year )year);
+			addChildElement(pElem, "maxdate", aDate.format_string("%F"));
+		}
 		addChildElement(pElem, "index", (queryIter->second.getIndexResults() ? "ALL" : "NONE"));
 		addChildElement(pElem, "label", queryIter->second.getLabelName());
 	}
@@ -1164,6 +1273,8 @@ bool PinotSettings::save(void)
 	}
 	// Ignore robots directives
 	addChildElement(pRootElem, "robots", (m_ignoreRobotsDirectives ? "IGNORE" : "OBEY"));
+	// Enable terms suggestion
+	addChildElement(pRootElem, "suggestterms", (m_suggestQueryTerms ? "YES" : "NO"));
 	// New results colour
 	pElem = pRootElem->add_child("newresults");
 	if (pElem == NULL)
@@ -1176,20 +1287,17 @@ bool PinotSettings::save(void)
 	addChildElement(pElem, "green", numStr);
 	sprintf(numStr, "%u", m_newResultsColourBlue);
 	addChildElement(pElem, "blue", numStr);
-	// Enable terms suggestion
-	addChildElement(pRootElem, "suggestterms", (m_suggestQueryTerms ? "YES" : "NO"));
-	// Mail accounts
-	for (set<TimestampedItem>::iterator mailIter = m_mailAccounts.begin(); mailIter != m_mailAccounts.end(); ++mailIter)
+	// Proxy
+	pElem = pRootElem->add_child("proxy");
+	if (pElem == NULL)
 	{
-		pElem = pRootElem->add_child("mailaccount");
-		if (pElem == NULL)
-		{
-			return false;
-		}
-		addChildElement(pElem, "name", mailIter->m_name);
-		sprintf(numStr, "%ld", mailIter->m_modTime);
-		addChildElement(pElem, "mtime", numStr);
+		return false;
 	}
+	addChildElement(pElem, "address", m_proxyAddress);
+	sprintf(numStr, "%u", m_proxyPort);
+	addChildElement(pElem, "port", numStr);
+	addChildElement(pElem, "type", m_proxyType);
+	addChildElement(pElem, "enable", (m_proxyEnabled ? "YES" : "NO"));
 	// Locations to index 
 	for (set<IndexableLocation>::iterator locationIter = m_indexableLocations.begin(); locationIter != m_indexableLocations.end(); ++locationIter)
 	{
@@ -1201,16 +1309,17 @@ bool PinotSettings::save(void)
 		addChildElement(pElem, "name", locationIter->m_name);
 		addChildElement(pElem, "monitor", (locationIter->m_monitor ? "YES" : "NO"));
 	}
-	// File patterns that are blacklisted
-	pElem = pRootElem->add_child("blacklist");
+	// File patterns
+	pElem = pRootElem->add_child("patterns");
 	if (pElem == NULL)
 	{
 		return false;
 	}
-	for (set<ustring>::iterator patternIter = m_filePatternsBlackList.begin(); patternIter != m_filePatternsBlackList.end() ; ++patternIter)
+	for (set<ustring>::iterator patternIter = m_filePatternsList.begin(); patternIter != m_filePatternsList.end() ; ++patternIter)
 	{
 		addChildElement(pElem, "pattern", *patternIter);
 	}
+	addChildElement(pElem, "forbid", (m_isBlackList ? "YES" : "NO"));
 
 	// Save to file
 	doc.write_to_file_formatted(getConfigurationFileName());
@@ -1339,7 +1448,7 @@ IndexInterface *PinotSettings::getIndex(const string &location)
 	{
 		return IndexFactory::getIndex("xapian", m_docsIndexLocation);
 	}
-	else if ((m_enableDBus == true) &&
+	else if ((m_clientMode == true) &&
 		(location == m_daemonIndexLocation))
 	{
 		return IndexFactory::getIndex("dbus", m_daemonIndexLocation);
@@ -1503,24 +1612,32 @@ void PinotSettings::clearLabels(void)
 /// Determines if a file matches the blacklist.
 bool PinotSettings::isBlackListed(const string &fileName)
 {
-	if (m_filePatternsBlackList.empty() == true)
+	if (m_filePatternsList.empty() == true)
 	{
-		return false;
+		if (m_isBlackList == true)
+		{
+			// There is no black-list
+			return false;
+		}
+
+		// The file is not in the (empty) whitelist
+		return true;
 	}
 
-	// Any pattern matches this the file name ?
-	for (set<ustring>::iterator patternIter = m_filePatternsBlackList.begin(); patternIter != m_filePatternsBlackList.end() ; ++patternIter)
+	// Any pattern matches this file name ?
+	for (set<ustring>::iterator patternIter = m_filePatternsList.begin(); patternIter != m_filePatternsList.end() ; ++patternIter)
 	{
 		if (fnmatch(patternIter->c_str(), fileName.c_str(), FNM_NOESCAPE) == 0)
 		{
 #ifdef DEBUG
 			cout << "PinotSettings::isBlackListed: " << fileName << " matches " << *patternIter << endl;
 #endif
-			return true;
+			// Fail if it's in the blacklist, let the file through otherwise
+			return m_isBlackList;
 		}
 	}
 
-	return false;
+	return !m_isBlackList;
 }
 
 PinotSettings::Engine::Engine()
