@@ -28,8 +28,10 @@
 #include <gtkmm/cellrendererprogress.h>
 
 #include "StringManip.h"
+#include "TimeConverter.h"
 #include "Url.h"
 #include "QueryHistory.h"
+#include "ResultsExporter.h"
 #include "ViewHistory.h"
 #include "config.h"
 #include "NLS.h"
@@ -44,17 +46,19 @@ using namespace Gdk;
 using namespace Gtk;
 
 ResultsTree::ResultsTree(const ustring &queryName, Menu *pPopupMenu,
-	PinotSettings &settings) :
+	GroupByMode groupMode, PinotSettings &settings) :
 	TreeView(),
-	m_queryName(queryName),
+	m_treeName(queryName),
 	m_pPopupMenu(pPopupMenu),
 	m_pResultsScrolledwindow(NULL),
 	m_settings(settings),
 	m_pExtractScrolledwindow(NULL),
 	m_extractTreeView(NULL),
 	m_showExtract(true),
-	m_groupBySearchEngine(true)
+	m_groupMode(groupMode)
 {
+	TreeViewColumn *pColumn = NULL;
+
 	m_pResultsScrolledwindow = manage(new ScrolledWindow());
 	m_pExtractScrolledwindow = manage(new ScrolledWindow());
 	m_extractTreeView = manage(new TreeView());
@@ -95,32 +99,35 @@ ResultsTree::ResultsTree(const ustring &queryName, Menu *pPopupMenu,
 	m_refStore = TreeStore::create(m_resultsColumns);
 	set_model(m_refStore);
 
-	// The first column is for the status icons
-	TreeViewColumn *pColumn = new TreeViewColumn("");
-	// Pack an icon renderer for the viewed status
-	CellRendererPixbuf *pIconRenderer = new CellRendererPixbuf();
-	pColumn->pack_start(*manage(pIconRenderer), false);
-	pColumn->set_cell_data_func(*pIconRenderer, SigC::slot(*this, &ResultsTree::renderViewStatus));
-	// Pack a second icon renderer for the indexed status
-	pIconRenderer = new CellRendererPixbuf();
-	pColumn->pack_start(*manage(pIconRenderer), false);
-	pColumn->set_cell_data_func(*pIconRenderer, SigC::slot(*this, &ResultsTree::renderIndexStatus));
-	// And a third one for the ranking
-	pIconRenderer = new CellRendererPixbuf();
-	pColumn->pack_start(*manage(pIconRenderer), false);
-	pColumn->set_cell_data_func(*pIconRenderer, SigC::slot(*this, &ResultsTree::renderRanking));
-	pColumn->set_sizing(TREE_VIEW_COLUMN_AUTOSIZE);
-	append_column(*manage(pColumn));
+	if (m_groupMode != FLAT)
+	{
+		// The first column is for the status icons
+		pColumn = new TreeViewColumn("");
+		// Pack an icon renderer for the viewed status
+		CellRendererPixbuf *pIconRenderer = new CellRendererPixbuf();
+		pColumn->pack_start(*manage(pIconRenderer), false);
+		pColumn->set_cell_data_func(*pIconRenderer, SigC::slot(*this, &ResultsTree::renderViewStatus));
+		// Pack a second icon renderer for the indexed status
+		pIconRenderer = new CellRendererPixbuf();
+		pColumn->pack_start(*manage(pIconRenderer), false);
+		pColumn->set_cell_data_func(*pIconRenderer, SigC::slot(*this, &ResultsTree::renderIndexStatus));
+		// And a third one for the ranking
+		pIconRenderer = new CellRendererPixbuf();
+		pColumn->pack_start(*manage(pIconRenderer), false);
+		pColumn->set_cell_data_func(*pIconRenderer, SigC::slot(*this, &ResultsTree::renderRanking));
+		pColumn->set_sizing(TREE_VIEW_COLUMN_AUTOSIZE);
+		append_column(*manage(pColumn));
 
-	// This is the score column
-	pColumn = new TreeViewColumn(_("Score"));
-	CellRendererProgress *pProgressRenderer = new CellRendererProgress();
-	pColumn->pack_start(*manage(pProgressRenderer));
-	pColumn->add_attribute(pProgressRenderer->property_text(), m_resultsColumns.m_scoreText);
-	pColumn->add_attribute(pProgressRenderer->property_value(), m_resultsColumns.m_score);
-	pColumn->set_sizing(TREE_VIEW_COLUMN_AUTOSIZE);
-	pColumn->set_sort_column(m_resultsColumns.m_score);
-	append_column(*manage(pColumn));
+		// This is the score column
+		pColumn = new TreeViewColumn(_("Score"));
+		CellRendererProgress *pProgressRenderer = new CellRendererProgress();
+		pColumn->pack_start(*manage(pProgressRenderer));
+		pColumn->add_attribute(pProgressRenderer->property_text(), m_resultsColumns.m_scoreText);
+		pColumn->add_attribute(pProgressRenderer->property_value(), m_resultsColumns.m_score);
+		pColumn->set_sizing(TREE_VIEW_COLUMN_AUTOSIZE);
+		pColumn->set_sort_column(m_resultsColumns.m_score);
+		append_column(*manage(pColumn));
+	}
 
 	// This is the title column
 	pColumn = new TreeViewColumn(_("Title"));
@@ -132,7 +139,7 @@ ResultsTree::ResultsTree(const ustring &queryName, Menu *pPopupMenu,
 	pColumn->set_sort_column(m_resultsColumns.m_text);
 	append_column(*manage(pColumn));
 
-	// The last column is for the URL
+	// URL
 	pColumn = create_column(_("URL"), m_resultsColumns.m_url, false, true, m_resultsColumns.m_url);
 	if (pColumn != NULL)
 	{
@@ -140,6 +147,14 @@ ResultsTree::ResultsTree(const ustring &queryName, Menu *pPopupMenu,
 		append_column(*manage(pColumn));
 	}
 
+	// The last column is for the timestamp
+	pColumn = create_column(_("Date"), m_resultsColumns.m_timestamp, false, true, m_resultsColumns.m_timestampTime);
+	if (pColumn != NULL)
+	{
+		pColumn->set_sizing(TREE_VIEW_COLUMN_AUTOSIZE);
+		append_column(*manage(pColumn));
+	}
+	
 	// Connect the signals
 	signal_button_press_event().connect_notify(
 		SigC::slot(*this, &ResultsTree::onButtonPressEvent));
@@ -291,9 +306,9 @@ void ResultsTree::renderTitleColumn(CellRenderer *pRenderer, const TreeModel::it
 			pTextRenderer->property_background_gdk().reset_value();
 		}
 
-		ResultsModelColumns::ResultType type = row[m_resultsColumns.m_type];
-		if ((type == ResultsModelColumns::RESULT_ROOT) ||
-			(type == ResultsModelColumns::RESULT_HOST))
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
+		if ((type == ResultsModelColumns::ROW_ENGINE) ||
+			(type == ResultsModelColumns::ROW_HOST))
 		{
 			ustring markup("<b>");
 			markup += row[m_resultsColumns.m_text];
@@ -343,13 +358,13 @@ void ResultsTree::onButtonPressEvent(GdkEventButton *ev)
 #ifdef DEBUG
 		cout << "ResultsTree::onButtonPressEvent: double click on button " << ev->button << endl;
 #endif
-		m_signalViewResults();
+		m_signalDoubleClick();
 	}
 }
 
 void ResultsTree::onSelectionChanged(void)
 {
-	m_signalSelectionChanged(m_queryName);
+	m_signalSelectionChanged(m_treeName);
 }
 
 bool ResultsTree::onSelectionSelect(const RefPtr<TreeModel>& model,
@@ -367,8 +382,8 @@ bool ResultsTree::onSelectionSelect(const RefPtr<TreeModel>& model,
 		m_refExtractStore->clear();
 
 		// Is this an actual result ?
-		ResultsModelColumns::ResultType type = row[m_resultsColumns.m_type];
-		if (type == ResultsModelColumns::RESULT_TITLE)
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
+		if (type == ResultsModelColumns::ROW_RESULT)
 		{
 #ifdef DEBUG
 			cout << "ResultsTree::onSelectionSelect: extract for " << row[m_resultsColumns.m_text] << endl;
@@ -442,24 +457,23 @@ ustring ResultsTree::getExtract(void) const
 // Adds a set of results.
 // Returns true if something was added to the tree.
 //
-bool ResultsTree::addResults(QueryProperties &queryProps, const string &engineName,
-	const vector<Result> &resultsList, const string &charset,
-	bool groupBySearchEngine)
+bool ResultsTree::addResults(const string &engineName, const vector<DocumentInfo> &resultsList,
+	const string &charset)
 {
 	std::map<string, TreeModel::iterator> updatedGroups;
-	ResultsModelColumns::ResultType rootType;
+	ResultsModelColumns::RowType rootType;
 	unsigned int count = 0;
 
 	// What's the grouping criteria ?
-	if (groupBySearchEngine == true)
+	if (m_groupMode == BY_ENGINE)
 	{
 		// By search engine
-		rootType = ResultsModelColumns::RESULT_ROOT;
+		rootType = ResultsModelColumns::ROW_ENGINE;
 	}
-	else
+	else if (m_groupMode == BY_HOST)
 	{
 		// By host
-		rootType = ResultsModelColumns::RESULT_HOST;
+		rootType = ResultsModelColumns::ROW_HOST;
 	}
 
 	// Find out what the search engine ID is
@@ -475,7 +489,7 @@ bool ResultsTree::addResults(QueryProperties &queryProps, const string &engineNa
 			indexId = m_settings.getIndexId(engineName);
 			engineId = m_settings.getEngineId("Xapian");
 #ifdef DEBUG
-			cout << "ResultsTree::addResults: engine is index " << engineName << endl;
+			cout << "ResultsTree::addResults: engine is index " << engineName << " " << indexId << endl;
 #endif
 		}
 #ifdef DEBUG
@@ -486,9 +500,9 @@ bool ResultsTree::addResults(QueryProperties &queryProps, const string &engineNa
 	cout << "ResultsTree::addResults: ID for engine " << engineName << " is " << engineId <<  endl;
 #endif
 
-	QueryHistory history(m_settings.m_historyDatabase);
+	QueryHistory history(m_settings.getHistoryDatabaseName());
 	bool isNewQuery = false;
-	if (history.getLastRun(m_queryName, engineName).empty() == true)
+	if (history.getLastRun(m_treeName, engineName).empty() == true)
 	{
 		isNewQuery = true;
 	}
@@ -497,49 +511,47 @@ bool ResultsTree::addResults(QueryProperties &queryProps, const string &engineNa
 #ifdef DEBUG
 	cout << "ResultsTree::addResults: " << resultsList.size() << " results" << endl;
 #endif
-	for (vector<Result>::const_iterator resultIter = resultsList.begin();
+	for (vector<DocumentInfo>::const_iterator resultIter = resultsList.begin();
 		resultIter != resultsList.end(); ++resultIter)
 	{
 		ustring title(to_utf8(resultIter->getTitle(), charset));
 		ustring location(to_utf8(resultIter->getLocation(), charset));
 		ustring extract(to_utf8(resultIter->getExtract(), charset));
+		string groupName;
+		TreeModel::iterator groupIter;
 		float currentScore = resultIter->getScore();
 		int rankDiff = 0;
 
-		// What group should the result go to ?
-		string groupName;
-		if (rootType == ResultsModelColumns::RESULT_HOST)
+		if (m_groupMode != FLAT)
 		{
-			Url urlObj(location);
-			groupName = urlObj.getHost();
-		}
-		else
-		{
-			groupName = engineName;
-		}
-
-		// Add the group or get its position if it's already in
-		TreeModel::iterator groupIter;
-		if (appendGroup(groupName, rootType, groupIter) == true)
-		{
-			// OK, add a row for this result within the group
-			TreeModel::iterator titleIter;
+			// What group should the result go to ?
+			if (rootType == ResultsModelColumns::ROW_HOST)
+			{
+				Url urlObj(location);
+				groupName = urlObj.getHost();
+			}
+			else
+			{
+				groupName = engineName;
+			}
+			// Add the group or get its position if it's already in
+			appendGroup(groupName, rootType, groupIter);
 
 			// Has the result's ranking changed ?
 			float oldestScore = 0;
-			float previousScore = history.hasItem(m_queryName, engineName,
+			float previousScore = history.hasItem(m_treeName, engineName,
 				location, oldestScore);
 			if (previousScore > 0)
 			{
 				// Update this result whatever the current and previous rankings were
-				history.updateItem(m_queryName, engineName, location,
+				history.updateItem(m_treeName, engineName, location,
 					title, extract, charset, currentScore);
 				rankDiff = (int)(currentScore - previousScore);
 			}
 			else
 			{
 				// No, this is a new result
-				history.insertItem(m_queryName, engineName, location,
+				history.insertItem(m_treeName, engineName, location,
 					resultIter->getTitle(), extract, charset, currentScore);
 				// New results are displayed as such only if the query has already been run on the engine
 				if (isNewQuery == false)
@@ -548,15 +560,31 @@ bool ResultsTree::addResults(QueryProperties &queryProps, const string &engineNa
 					rankDiff = 666;
 				}
 			}
+		}
 
-			++count;
-			if (appendResult(title, location, (int)currentScore, rankDiff,
-				engineId, indexId, titleIter, &(*groupIter), true) == true)
-			{
+		++count;
+
+		unsigned int docId = resultIter->getIsIndexed(indexId);
+		bool isIndexed = false;
+
+		if (docId > 0)
+		{
+			isIndexed = true;
+		}
+
+		// OK, add a row for this result within the group
+		TreeModel::iterator titleIter;
+		if (appendResult(title, location, resultIter->getType(), (int)currentScore, rankDiff, isIndexed,
+			docId, resultIter->getTimestamp(), engineId, indexId, titleIter, groupIter, true) == true)
+		{
 #ifdef DEBUG
-				cout << "ResultsTree::addResults: added row for result " << count << ", " << currentScore << endl;
+			cout << "ResultsTree::addResults: added row for result " << count
+				<< ", " << currentScore << ", " << isIndexed << " " << indexId
+				<< " " << resultIter->getTimestamp() << endl;
 #endif
 
+			if (groupIter)
+			{
 				// Update this map, so that we know which groups need updating
 				updatedGroups[groupName] = groupIter;
 			}
@@ -578,17 +606,14 @@ bool ResultsTree::addResults(QueryProperties &queryProps, const string &engineNa
 
 		return true;
 	}
-	else
+	else if (m_groupMode == BY_ENGINE)
 	{
-		if (rootType == ResultsModelColumns::RESULT_ROOT)
-		{
-			// If this didn't return any result, add an empty group
-			TreeModel::iterator groupIter;
-			appendGroup(engineName, rootType, groupIter);
-			updateGroup(groupIter);
+		// If this didn't return any result, add an empty group
+		TreeModel::iterator groupIter;
+		appendGroup(engineName, rootType, groupIter);
+		updateGroup(groupIter);
 
-			return true;
-		}
+		return true;
 	}
 
 	return false;
@@ -597,16 +622,25 @@ bool ResultsTree::addResults(QueryProperties &queryProps, const string &engineNa
 //
 // Sets how results are grouped.
 //
-void ResultsTree::setGroupMode(bool groupBySearchEngine)
+void ResultsTree::setGroupMode(GroupByMode groupMode)
 {
-	ResultsModelColumns::ResultType currentType, newType;
+	ResultsModelColumns::RowType currentType, newType;
 
-	if (m_groupBySearchEngine == groupBySearchEngine)
+	if (m_groupMode == FLAT)
+	{
+		// No change possible
+		return;
+	}
+
+	if (m_groupMode == groupMode)
 	{
 		// No change
 		return;
 	}
-	m_groupBySearchEngine = groupBySearchEngine;
+#ifdef DEBUG
+	cout << "ResultsTree::setGroupMode: set to " << groupMode << endl;
+#endif
+	m_groupMode = groupMode;
 
 	// Do we need to update the tree ?
 	TreeModel::Children children = m_refStore->children();
@@ -616,17 +650,17 @@ void ResultsTree::setGroupMode(bool groupBySearchEngine)
 	}
 
 	// What's the new grouping criteria ?
-	if (groupBySearchEngine == true)
+	if (m_groupMode == BY_ENGINE)
 	{
 		// By search engine
-		currentType = ResultsModelColumns::RESULT_HOST;
-		newType = ResultsModelColumns::RESULT_ROOT;
+		currentType = ResultsModelColumns::ROW_HOST;
+		newType = ResultsModelColumns::ROW_ENGINE;
 	}
 	else
 	{
 		// By host
-		currentType = ResultsModelColumns::RESULT_ROOT;
-		newType = ResultsModelColumns::RESULT_HOST;
+		currentType = ResultsModelColumns::ROW_ENGINE;
+		newType = ResultsModelColumns::ROW_HOST;
 	}
 
 	// Clear the map
@@ -642,7 +676,7 @@ void ResultsTree::setGroupMode(bool groupBySearchEngine)
 #ifdef DEBUG
 		cout << "ResultsTree::setGroupMode: looking at " << row[m_resultsColumns.m_text] << endl;
 #endif
-		ResultsModelColumns::ResultType type = row[m_resultsColumns.m_type];
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
 		// Skip new type rows
 		if (type == newType)
 		{
@@ -650,121 +684,137 @@ void ResultsTree::setGroupMode(bool groupBySearchEngine)
 			continue;
 		}
 
-		TreeModel::Children child = iter->children();
-		if (child.empty() == false)
+		TreeModel::Children groupChildren = iter->children();
+		TreeModel::Children::iterator childIter = groupChildren.begin();
+		// Type ROW_RESULT
+		while (childIter != groupChildren.end())
 		{
-			TreeModel::Children::iterator childIter = child.begin();
-			// Type RESULT_TITLE
-			while (childIter != child.end())
+			TreeModel::Row childRow = *childIter;
+			TreeModel::iterator groupIter, newIter;
+			bool success = false;
+
+			// We will need the URL and engines columns in all cases
+			string url(from_utf8(childRow[m_resultsColumns.m_url]));
+			unsigned int engineIds = childRow[m_resultsColumns.m_engines];
+			unsigned int indexIds = childRow[m_resultsColumns.m_indexes];
+
+			// Get the name of the group this should go into
+			if (m_groupMode == BY_HOST)
 			{
-				TreeModel::Row childRow = *childIter;
-				TreeModel::iterator groupIter, newIter;
-				bool success = false;
-
-				// We will need the URL and engines columns in all cases
-				string url = from_utf8(childRow[m_resultsColumns.m_url]);
-				unsigned int engineIds = childRow[m_resultsColumns.m_engines];
-				unsigned int indexIds = childRow[m_resultsColumns.m_indexes];
-
-				// Get the name of the group this should go into
-				if (newType == ResultsModelColumns::RESULT_HOST)
-				{
-					Url urlObj(url);
+				Url urlObj(url);
 #ifdef DEBUG
-					cout << "ResultsTree::setGroupMode: row " << url << endl;
+				cout << "ResultsTree::setGroupMode: row " << url << endl;
 #endif
-					string groupName = urlObj.getHost();
-					// Add group
-					if (appendGroup(groupName, newType, groupIter) == true)
+				// Add group
+				if (appendGroup(urlObj.getHost(), newType, groupIter) == true)
+				{
+					// Add result
+					success = appendResult(childRow[m_resultsColumns.m_text],
+						childRow[m_resultsColumns.m_url],
+						childRow[m_resultsColumns.m_type],
+						childRow[m_resultsColumns.m_score],
+						childRow[m_resultsColumns.m_rankDiff],
+						childRow[m_resultsColumns.m_indexed],
+						childRow[m_resultsColumns.m_docId],
+						childRow[m_resultsColumns.m_timestamp],
+						engineIds, indexIds,
+						newIter, groupIter, true);
+				}
+			}
+			else
+			{
+				// Look at the engines column and see which engines this result is for
+				set<string> engineNames;
+				m_settings.getEngineNames(engineIds, engineNames);
+				if (engineNames.empty() == false)
+				{
+#ifdef DEBUG
+					cout << "ResultsTree::setGroupMode: row is for " << engineNames.size() << " engine(s)" << endl;
+#endif
+					// Are there indexes in the list ?
+					set<string>::iterator xapianIter = engineNames.find("Xapian");
+					if ((xapianIter != engineNames.end()) &&
+						(indexIds > 0))
 					{
-						// Add result
-						success = appendResult(childRow[m_resultsColumns.m_text],
-							childRow[m_resultsColumns.m_url],
-							childRow[m_resultsColumns.m_score],
-							childRow[m_resultsColumns.m_rankDiff],
-							engineIds, indexIds, newIter, &(*groupIter), true);
+						// Erase this
+						engineNames.erase(xapianIter);
+#ifdef DEBUG
+						cout << "ResultsTree::setGroupMode: row is for index(es)" << indexIds << endl;
+#endif
+
+						// Add entries for each index name so that we can loop once on engine names
+						set<string> indexNames;
+						m_settings.getIndexNames(indexIds, indexNames);
+						for (set<string>::iterator iter = indexNames.begin();
+							iter != indexNames.end(); ++iter)
+						{
+							string indexName(*iter);
+							engineNames.insert(indexName);
+#ifdef DEBUG
+							cout << "ResultsTree::setGroupMode: row is for index " << indexName << endl;
+#endif
+						}
 					}
-				}
-				else
-				{
-					// Look at the engines column and see which engines this result is for
-					set<string> engineNames;
-					m_settings.getEngineNames(engineIds, engineNames);
-					if (engineNames.empty() == false)
+
+					for (set<string>::iterator iter = engineNames.begin();
+						iter != engineNames.end(); ++iter)
 					{
-#ifdef DEBUG
-						cout << "ResultsTree::setGroupMode: row is for " << engineNames.size() << endl;
-#endif
-						// Are there indexes in the list ?
-						set<string>::iterator xapianIter = engineNames.find("Xapian");
-						if ((xapianIter != engineNames.end()) &&
-							(indexIds > 0))
+						string engineName(*iter);
+						unsigned int indexId = 0;
+						unsigned int engineId = m_settings.getEngineId(engineName);
+
+						if (engineId == 0)
 						{
-							// Erase this
-							engineNames.erase(xapianIter);
-
-							// Add entries for each index name so that we can loop once on engine names
-							set<string> indexNames;
-							m_settings.getIndexNames(indexIds, indexNames);
-							for (set<string>::iterator iter = indexNames.begin(); iter != indexNames.end(); ++iter)
+							// This is actually an index, not an engine...
+							indexId = m_settings.getIndexId(engineName);
+							if (indexId > 0)
 							{
-								string indexName = (*iter);
-								engineNames.insert(indexName);
-#ifdef DEBUG
-								cout << "ResultsTree::setGroupMode: row is for index " << indexName << endl;
-#endif
+								engineId = m_settings.getEngineId("Xapian");
 							}
+#ifdef DEBUG
+							cout << "ResultsTree::setGroupMode: index " << indexId << endl;
+#endif
 						}
+#ifdef DEBUG
+						else cout << "ResultsTree::setGroupMode: no index" << endl;
+#endif
 
-						for (set<string>::iterator iter = engineNames.begin(); iter != engineNames.end(); ++iter)
+						// Add group
+						if (appendGroup(engineName, newType, groupIter) == true)
 						{
-							string engineName = (*iter);
-							unsigned int indexId = 0;
-							unsigned int engineId = m_settings.getEngineId(engineName);
-
-							if (engineId == 0)
-							{
-								// This is actually an index, not an engine...
-								indexId = m_settings.getIndexId(engineName);
-								if (indexId > 0)
-								{
-									engineId = m_settings.getEngineId("Xapian");
-								}
-							}
-
-							// Add group
-							if (appendGroup(engineName, newType, groupIter) == true)
-							{
-								// Add result
-								appendResult(childRow[m_resultsColumns.m_text],
-									childRow[m_resultsColumns.m_url],
-									childRow[m_resultsColumns.m_score],
-									childRow[m_resultsColumns.m_rankDiff],
-									engineId, indexId,
-									newIter, &(*groupIter), true);
+							// Add result
+							appendResult(childRow[m_resultsColumns.m_text],
+								childRow[m_resultsColumns.m_url],
+								childRow[m_resultsColumns.m_type],
+								childRow[m_resultsColumns.m_score],
+								childRow[m_resultsColumns.m_rankDiff],
+								childRow[m_resultsColumns.m_indexed],
+								childRow[m_resultsColumns.m_docId],
+								childRow[m_resultsColumns.m_timestamp],
+								engineId, indexId,
+								newIter, groupIter, true);
 #ifdef DEBUG
-								cout << "ResultsTree::setGroupMode: row for " << *iter << endl;
+							cout << "ResultsTree::setGroupMode: row for " << *iter << endl;
 #endif
-							}
 						}
-
-						// FIXME: make sure at least one row was added
-						success = true;
 					}
-				}
 
-				if (success == true)
-				{
-					// Delete it
-					m_refStore->erase(*childIter);
-					childIter = child.begin();
-				}
-				else
-				{
-					// Don't delete anything then, just go to the next child
-					childIter++;
+					// FIXME: make sure at least one row was added
+					success = true;
 				}
 			}
+
+			if (success == true)
+			{
+				// Delete it
+				m_refStore->erase(*childIter);
+				childIter = groupChildren.begin();
+			}
+			else
+			{
+				// Don't delete anything then, just go to the next child
+				childIter++;
+			}
 		}
 
 		// Erase this row
@@ -784,10 +834,28 @@ void ResultsTree::setGroupMode(bool groupBySearchEngine)
 	onSelectionChanged();
 }
 
+//
+// Gets the first selected item's URL.
+//
+ustring ResultsTree::getFirstSelectionURL(void)
+{
+	list<TreeModel::Path> selectedItems = get_selection()->get_selected_rows();
+	if (selectedItems.empty() == true)
+	{
+		return "";
+	}
+
+	list<TreeModel::Path>::iterator itemPath = selectedItems.begin();
+	TreeModel::iterator iter = m_refStore->get_iter(*itemPath);
+	TreeModel::Row row = *iter;
+
+	return row[m_resultsColumns.m_url];
+}
+
 //
 // Gets a list of selected items.
 //
-bool ResultsTree::getSelection(vector<DocumentInfo> &resultsList)
+bool ResultsTree::getSelection(vector<DocumentInfo> &resultsList, bool skipIndexed)
 {
 	list<TreeModel::Path> selectedItems = get_selection()->get_selected_rows();
 	if (selectedItems.empty() == true)
@@ -802,10 +870,41 @@ bool ResultsTree::getSelection(vector<DocumentInfo> &resultsList)
 		TreeModel::iterator iter = m_refStore->get_iter(*itemPath);
 		TreeModel::Row row = *iter;
 
-		if (row[m_resultsColumns.m_type] == ResultsModelColumns::RESULT_TITLE)
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
+		if (type != ResultsModelColumns::ROW_RESULT)
 		{
-			resultsList.push_back(DocumentInfo(from_utf8(row[m_resultsColumns.m_text]),
-				from_utf8(row[m_resultsColumns.m_url]), "", ""));
+			continue;
+		}
+
+		bool isIndexed = row[m_resultsColumns.m_indexed];
+		if ((skipIndexed == false) ||
+			(isIndexed == false))
+		{
+			DocumentInfo current(from_utf8(row[m_resultsColumns.m_text]),
+				from_utf8(row[m_resultsColumns.m_url]),
+				from_utf8(row[m_resultsColumns.m_type]), "");
+
+			if (isIndexed == true)
+			{
+				set<string> indexNames;
+				unsigned int indexIds = row[m_resultsColumns.m_indexes];
+
+				m_settings.getIndexNames(indexIds, indexNames);
+				// Any internal index in there ?
+				for (set<string>::iterator indexIter = indexNames.begin(); indexIter != indexNames.end(); ++indexIter)
+				{
+					if  (m_settings.isInternalIndex(*indexIter) == true)
+					{
+#ifdef DEBUG
+						cout << "ResultsTree::getSelection: result in internal index " << *indexIter << endl;
+#endif
+						current.setIsIndexed(m_settings.getIndexId(*indexIter), row[m_resultsColumns.m_docId]);
+						break;
+					}
+				}
+			}
+
+			resultsList.push_back(current);
 		}
 	}
 #ifdef DEBUG
@@ -818,7 +917,7 @@ bool ResultsTree::getSelection(vector<DocumentInfo> &resultsList)
 //
 // Sets the selected items' state.
 //
-void ResultsTree::setSelectionState(bool viewed, bool indexed)
+void ResultsTree::setSelectionState(bool viewed)
 {
 	list<TreeModel::Path> selectedItems = get_selection()->get_selected_rows();
 	if (selectedItems.empty() == true)
@@ -832,8 +931,9 @@ void ResultsTree::setSelectionState(bool viewed, bool indexed)
 	{
 		TreeModel::iterator iter = m_refStore->get_iter(*itemPath);
 		TreeModel::Row row = *iter;
-
-		if (row[m_resultsColumns.m_type] != ResultsModelColumns::RESULT_TITLE)
+  
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
+		if (type != ResultsModelColumns::ROW_RESULT)
 		{
 			continue;
 		}
@@ -842,9 +942,37 @@ void ResultsTree::setSelectionState(bool viewed, bool indexed)
 		{
 			row[m_resultsColumns.m_viewed] = true;
 		}
-		if (indexed == true)
+	}
+}
+
+//
+// Updates a result's properties.
+//
+void ResultsTree::updateResult(const DocumentInfo &result)
+{
+	unsigned int indexId = 0;
+	unsigned int docId = result.getIsIndexed(indexId);
+
+	if (docId == 0)
+	{
+		return;
+	}
+
+	// Go through the list
+	TreeModel::Children children = m_refStore->children();
+	for (TreeModel::Children::iterator iter = children.begin(); iter != children.end(); ++iter)
+	{
+		TreeModel::Row row = *iter;
+
+		if (docId == row[m_resultsColumns.m_docId])
 		{
-			row[m_resultsColumns.m_indexed] = true;
+			updateRow(row, result.getTitle(), result.getLocation(), result.getType(),
+				row[m_resultsColumns.m_score], row[m_resultsColumns.m_engines],
+				row[m_resultsColumns.m_indexes], docId,
+				result.getTimestamp(), ResultsModelColumns::ROW_RESULT,
+				row[m_resultsColumns.m_indexed], row[m_resultsColumns.m_viewed],
+				row[m_resultsColumns.m_rankDiff]);
+			break;
 		}
 	}
 }
@@ -867,9 +995,10 @@ bool ResultsTree::deleteSelection(void)
 		bool updateParent = false;
 
 		// This could be a group that's in the map and should be removed first
-		if (row[m_resultsColumns.m_type] != ResultsModelColumns::RESULT_TITLE)
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
+		if (type != ResultsModelColumns::ROW_RESULT)
 		{
-			string groupName = from_utf8(row[m_resultsColumns.m_text]);
+			string groupName(from_utf8(row[m_resultsColumns.m_text]));
 			std::map<string, TreeModel::iterator>::iterator mapIter = m_resultsGroups.find(groupName);
 			if (mapIter != m_resultsGroups.end())
 			{
@@ -879,7 +1008,7 @@ bool ResultsTree::deleteSelection(void)
 #endif
 			}
 		}
-		else
+		else if (m_groupMode != FLAT)
 		{
 			// This item is a result
 			parentIter = row.parent();
@@ -912,9 +1041,8 @@ bool ResultsTree::deleteSelection(void)
 //
 // Deletes results.
 //
-bool ResultsTree::deleteResults(QueryProperties &queryProps, const string &engineName)
+bool ResultsTree::deleteResults(const string &engineName)
 {
-	ustring queryName(queryProps.getName());
 	unsigned int indexId = 0;
 	unsigned int engineId = m_settings.getEngineId(engineName);
 	unsigned int count = 0;
@@ -937,8 +1065,9 @@ bool ResultsTree::deleteResults(QueryProperties &queryProps, const string &engin
 	{
 		TreeModel::Row row = *parentIter;
 
-		if ((row[m_resultsColumns.m_type] != ResultsModelColumns::RESULT_ROOT) &&
-			(row[m_resultsColumns.m_type] != ResultsModelColumns::RESULT_HOST))
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
+		if ((type != ResultsModelColumns::ROW_ENGINE) &&
+			(type != ResultsModelColumns::ROW_HOST))
 		{
 			continue;
 		}
@@ -949,7 +1078,8 @@ bool ResultsTree::deleteResults(QueryProperties &queryProps, const string &engin
 		{
 			row = *iter;
 
-			if ((row[m_resultsColumns.m_type] == ResultsModelColumns::RESULT_TITLE) &&
+			type = row[m_resultsColumns.m_resultType];
+			if ((type == ResultsModelColumns::ROW_RESULT) &&
 				(row[m_resultsColumns.m_engines] == engineId) &&
 				(row[m_resultsColumns.m_indexes] == indexId))
 			{
@@ -980,6 +1110,14 @@ bool ResultsTree::deleteResults(QueryProperties &queryProps, const string &engin
 	return false;
 }
 
+//
+// Returns the number of rows.
+//
+unsigned int ResultsTree::getRowsCount(void)
+{
+	return m_refStore->children().size();
+}
+
 //
 // Refreshes the tree.
 //
@@ -1042,6 +1180,113 @@ void ResultsTree::showExtract(bool show)
 	}
 }
 
+//
+// Exports results to a file.
+//
+void ResultsTree::exportResults(const string &fileName, bool csvFormat)
+{
+	QueryHistory history(m_settings.getHistoryDatabaseName());
+	QueryProperties queryProps(m_treeName, "");
+	ResultsExporter *pExporter = NULL;
+	unsigned int maxResultsCount = 0;
+
+	if (fileName.empty() == true)
+	{
+		return;
+	}
+
+	if (csvFormat == true)
+	{
+		pExporter = new CSVExporter(fileName,
+			queryProps);
+	}
+	else
+	{
+		pExporter = new OpenSearchExporter(fileName,
+			queryProps);
+	}
+
+	// How many results are there altogether ?
+	TreeModel::Children children = m_refStore->children();
+	for (TreeModel::Children::iterator iter = children.begin();
+		iter != children.end(); ++iter)
+	{
+		TreeModel::Row row = *iter;
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
+
+		if ((type != ResultsModelColumns::ROW_ENGINE) &&
+			(type != ResultsModelColumns::ROW_HOST))
+		{
+			continue;
+		}
+
+		TreeModel::Children groupChildren = iter->children();
+		maxResultsCount += groupChildren.size();
+	}
+
+	// Start
+	pExporter->exportStart("", maxResultsCount);
+
+	for (TreeModel::Children::iterator iter = children.begin();
+		iter != children.end(); ++iter)
+	{
+		TreeModel::Row row = *iter;
+		ResultsModelColumns::RowType type = row[m_resultsColumns.m_resultType];
+
+		if ((type != ResultsModelColumns::ROW_ENGINE) &&
+			(type != ResultsModelColumns::ROW_HOST))
+		{
+			continue;
+		}
+
+		TreeModel::Children groupChildren = iter->children();
+		for (TreeModel::Children::iterator childIter = groupChildren.begin();
+			childIter != groupChildren.end(); ++childIter)
+		{
+			set<string> engineNames, indexNames;
+			TreeModel::Row row = *childIter;
+			DocumentInfo result(from_utf8(row[m_resultsColumns.m_text]),
+				from_utf8(row[m_resultsColumns.m_url]),
+				from_utf8(row[m_resultsColumns.m_type]), "");
+			string engineName, charset;
+			unsigned int engineIds = row[m_resultsColumns.m_engines];
+			unsigned int indexIds = row[m_resultsColumns.m_indexes];
+
+#ifdef DEBUG
+			cout << "ResultsTree::exportResults: engines " << engineIds << ", indexes " << indexIds << endl;
+#endif
+			m_settings.getEngineNames(engineIds, engineNames);
+			if (engineNames.empty() == false)
+			{
+				// Get the first engine this result was obtained from
+				engineName = *engineNames.begin();
+				if (engineName == "Xapian")
+				{
+					m_settings.getIndexNames(indexIds, indexNames);
+					if (indexNames.empty() == false)
+					{
+						// Use the name of the first index as engine name
+						engineName = (*indexNames.begin());
+					}
+				}
+			}
+			if (m_groupMode != FLAT)
+			{
+				result.setExtract(history.getItemExtract(from_utf8(m_treeName),
+					engineName, result.getLocation(), charset));
+			}
+
+			// Export this
+			pExporter->exportResult(engineName, result);
+		}
+	}
+
+	// End
+	pExporter->exportEnd();
+
+	delete pExporter;
+}
+
 //
 // Returns the changed selection signal.
 //
@@ -1051,33 +1296,40 @@ Signal1<void, ustring>& ResultsTree::getSelectionChangedSignal(void)
 }
 
 //
-// Returns the view results signal.
+// Returns the double-click signal.
 //
-Signal0<void>& ResultsTree::getViewResultsSignal(void)
+Signal0<void>& ResultsTree::getDoubleClickSignal(void)
 {
-	return m_signalViewResults;
+	return m_signalDoubleClick;
 }
 
 //
 // Adds a new row in the results tree.
 //
 bool ResultsTree::appendResult(const ustring &text, const ustring &url,
-	int score, int rankDiff, unsigned int engineId, unsigned int indexId,
-	TreeModel::iterator &newRowIter, const TreeModel::Row *parentRow, bool noDuplicates)
+	const ustring &type, int score, int rankDiff, bool isIndexed,
+	unsigned int docId, const ustring &timestamp, unsigned int engineId,
+	unsigned int indexId, TreeModel::iterator &newRowIter,
+	const TreeModel::iterator &parentIter, bool noDuplicates)
 {
-	if (parentRow == NULL)
+	if (!parentIter)
 	{
+#ifdef DEBUG
+		cout << "ResultsTree::appendResult: no parent" << endl;
+#endif
 		newRowIter = m_refStore->append();
 	}
 	else
 	{
+		const TreeModel::Row parentRow = *parentIter;
+
 		// Merge duplicates within groups ?
 		if (noDuplicates == true)
 		{
 			// Look for a row with the same URL and query. For instance, in group
 			// by host mode, if a page is returned by several search engines, it
 			// should appear only once
-			TreeModel::Children children = parentRow->children();
+			TreeModel::Children children = parentRow.children();
 			if (children.empty() == false)
 			{
 				for (TreeModel::Children::iterator childIter = children.begin();
@@ -1098,30 +1350,16 @@ bool ResultsTree::appendResult(const ustring &text, const ustring &url,
 			}
 		}
 
-		newRowIter = m_refStore->append(parentRow->children());
-	}
-
-	IndexInterface *pIndex = m_settings.getIndex("MERGED");
-	ViewHistory viewHistory(m_settings.m_historyDatabase);
-	bool isIndexed = false;
-
-	// Is this document indexed ?
-	if (pIndex != NULL)
-	{
-		if (pIndex->hasDocument(url) > 0)
-		{
-			isIndexed = true;
-		}
-
-		delete pIndex;
+		newRowIter = m_refStore->append(parentRow.children());
 	}
 
 	// Has it been already viewed ?
+	ViewHistory viewHistory(m_settings.getHistoryDatabaseName());
 	bool wasViewed = viewHistory.hasItem(url);
 
 	TreeModel::Row childRow = *newRowIter;
-	updateRow(childRow, text, url, score, engineId, indexId,
-		ResultsModelColumns::RESULT_TITLE, isIndexed,
+	updateRow(childRow, text, url, type, score, engineId, indexId,
+		docId, timestamp, ResultsModelColumns::ROW_RESULT, isIndexed,
 		wasViewed, rankDiff);
 
 	return true;
@@ -1130,7 +1368,7 @@ bool ResultsTree::appendResult(const ustring &text, const ustring &url,
 //
 // Adds a results group
 //
-bool ResultsTree::appendGroup(const string &groupName, ResultsModelColumns::ResultType groupType,
+bool ResultsTree::appendGroup(const string &groupName, ResultsModelColumns::RowType groupType,
 	TreeModel::iterator &groupIter)
 {
 	bool success = false;
@@ -1143,7 +1381,7 @@ bool ResultsTree::appendGroup(const string &groupName, ResultsModelColumns::Resu
 		groupIter = m_refStore->append();
 		TreeModel::Row groupRow = *groupIter;
 		updateRow(groupRow, to_utf8(groupName),
-			"", 0, 0, 0, groupType,
+			"", "", 0, 0, 0, 0, "", groupType,
 			false, false, false);
 
 		// Update the map
@@ -1175,7 +1413,8 @@ void ResultsTree::updateGroup(TreeModel::iterator &groupIter)
 	int averageScore = 0;
 
 	// Check the iterator doesn't point to a result
-	if (groupRow[m_resultsColumns.m_type] == ResultsModelColumns::RESULT_TITLE)
+	ResultsModelColumns::RowType type = groupRow[m_resultsColumns.m_resultType];
+	if (type == ResultsModelColumns::ROW_RESULT)
 	{
 		return;
 	}
@@ -1210,18 +1449,24 @@ void ResultsTree::updateGroup(TreeModel::iterator &groupIter)
 // Updates a row.
 //
 void ResultsTree::updateRow(TreeModel::Row &row, const ustring &text,
-	const ustring &url, int score, unsigned int engineId,  unsigned int indexId,
-	ResultsModelColumns::ResultType type, bool indexed, bool viewed, int rankDiff)
+	const ustring &url, const ustring &type, int score,
+	unsigned int engineId, unsigned int indexId,
+	unsigned int docId, const ustring &timestamp,
+	ResultsModelColumns::RowType resultType, bool indexed, bool viewed, int rankDiff)
 {
 	try
 	{
 		row[m_resultsColumns.m_text] = text;
 		row[m_resultsColumns.m_url] = url;
+		row[m_resultsColumns.m_type] = type;
 		row[m_resultsColumns.m_score] = score;
 		row[m_resultsColumns.m_scoreText] = "";
 		row[m_resultsColumns.m_engines] = engineId;
 		row[m_resultsColumns.m_indexes] = indexId;
-		row[m_resultsColumns.m_type] = type;
+		row[m_resultsColumns.m_docId] = docId;
+		row[m_resultsColumns.m_resultType] = resultType;
+		row[m_resultsColumns.m_timestamp] = timestamp;
+		row[m_resultsColumns.m_timestampTime] = TimeConverter::fromTimestamp(timestamp);
 
 		row[m_resultsColumns.m_indexed] = indexed;
 		row[m_resultsColumns.m_viewed] = viewed;
@@ -1229,11 +1474,15 @@ void ResultsTree::updateRow(TreeModel::Row &row, const ustring &text,
 	}
 	catch (Error &error)
 	{
-		cerr << "ResultsTree::updateRow: " << error.what() << endl;
+#ifdef DEBUG
+		cout << "ResultsTree::updateRow: " << error.what() << endl;
+#endif
 	}
 	catch (...)
 	{
-		cerr << "ResultsTree::updateRow: caught unknown exception" << endl;
+#ifdef DEBUG
+		cout << "ResultsTree::updateRow: caught unknown exception" << endl;
+#endif
 	}
 }
 
@@ -1242,7 +1491,7 @@ void ResultsTree::updateRow(TreeModel::Row &row, const ustring &text,
 //
 ustring ResultsTree::findResultsExtract(const Gtk::TreeModel::Row &row)
 {
-	QueryHistory history(m_settings.m_historyDatabase);
+	QueryHistory history(m_settings.getHistoryDatabaseName());
 	set<string> engineNames, indexNames;
 	string url(from_utf8(row[m_resultsColumns.m_url]));
 	string extract, charset;
@@ -1256,7 +1505,7 @@ ustring ResultsTree::findResultsExtract(const Gtk::TreeModel::Row &row)
 	if (engineNames.empty() == false)
 	{
 		// Get the first engine this result was obtained from
-		string engineName = (*engineNames.begin());
+		string engineName(*engineNames.begin());
 		if (engineName == "Xapian")
 		{
 			m_settings.getIndexNames(indexIds, indexNames);
@@ -1279,7 +1528,7 @@ ustring ResultsTree::findResultsExtract(const Gtk::TreeModel::Row &row)
 #ifdef DEBUG
 		cout << "ResultsTree::findResultsExtract: first engine for " << url << " was " << engineName << endl;
 #endif
-		extract = history.getItemExtract(from_utf8(m_queryName), engineName, url, charset);
+		extract = history.getItemExtract(from_utf8(m_treeName), engineName, url, charset);
 	}
 
 	return to_utf8(extract, charset);
