@@ -27,15 +27,18 @@
 #include <errno.h>
 #include <exception>
 #include <iostream>
+#include <fstream>
 #include <sigc++/class_slot.h>
 #include <sigc++/compatibility.h>
 #include <sigc++/slot.h>
 #include <glibmm/miscutils.h>
 
 #include "Languages.h"
+#include "MIMEScanner.h"
 #include "TimeConverter.h"
+#include "Timer.h"
 #include "Url.h"
-#include "CrawlHistory.h"
+#include "DBusXapianIndex.h"
 #include "XapianIndex.h"
 #include "XapianEngine.h"
 #include "config.h"
@@ -64,6 +67,47 @@ static DBusMessage *newDBusReply(DBusMessage *pMessage)
         return NULL;
 }
 
+static ustring g_xmlDescription;
+
+static bool loadXMLDescription(void)
+{
+	bool readFile = false;
+
+	if (g_xmlDescription.empty() == false)
+	{
+		return true;
+	}
+
+	ifstream xmlFile;
+	string xmlFileName(PREFIX);
+
+	xmlFileName += "/share/pinot/pinot-dbus-daemon.xml";
+	xmlFile.open(xmlFileName.c_str());
+	if (xmlFile.good() == true)
+	{
+		xmlFile.seekg(0, ios::end);
+		int length = xmlFile.tellg();
+		xmlFile.seekg(0, ios::beg);
+
+		char *pXmlBuffer = new char[length + 1];
+		xmlFile.read(pXmlBuffer, length);
+		if (xmlFile.fail() == false)
+		{
+			pXmlBuffer[length] = '\0';
+			g_xmlDescription = pXmlBuffer;
+			readFile = true;
+		}
+		delete[] pXmlBuffer;
+	}
+	xmlFile.close();
+
+	if (readFile == false)
+	{
+		cerr << "File " << xmlFileName << " couldn't be read" << endl;
+	}
+	return readFile;
+}
+
 MonitorThread::MonitorThread(MonitorInterface *pMonitor, MonitorHandler *pHandler) :
 	WorkerThread(),
 	m_ctrlReadPipe(-1),
@@ -118,7 +162,7 @@ Signal3<void, const DocumentInfo&, const std::string&, bool>& MonitorThread::get
 
 void MonitorThread::processEvents(void)
 {
-	CrawlHistory history(PinotSettings::getInstance().m_historyDatabase);
+	CrawlHistory history(PinotSettings::getInstance().getHistoryDatabaseName());
 	queue<MonitorEvent> events;
 
 #ifdef DEBUG
@@ -153,6 +197,16 @@ void MonitorThread::processEvents(void)
 			<< event.m_location << " " << event.m_isDirectory << endl;
 #endif
 
+		// Skip dotfiles and blacklisted files
+		Url urlObj("file://" + event.m_location);
+		if ((urlObj.getFile()[0] == '.') ||
+			(PinotSettings::getInstance().isBlackListed(event.m_location) == true))
+		{
+			// Next
+			events.pop();
+			continue;
+		}
+
 		// What's the event code ?
 		if (event.m_type == MonitorEvent::EXISTS)
 		{
@@ -299,20 +353,21 @@ void MonitorThread::doWork(void)
 }
 
 DirectoryScannerThread::DirectoryScannerThread(const string &dirName, bool isSource,
-	unsigned int maxLevel, bool followSymLinks,
-	MonitorInterface *pMonitor, MonitorHandler *pHandler) :
+	bool fullScan, MonitorInterface *pMonitor, MonitorHandler *pHandler,
+	unsigned int maxLevel, bool followSymLinks) :
 	WorkerThread(),
 	m_dirName(dirName),
-	m_maxLevel(maxLevel),
-	m_followSymLinks(followSymLinks),
+	m_fullScan(fullScan),
 	m_pMonitor(pMonitor),
 	m_pHandler(pHandler),
+	m_sourceId(0),
 	m_currentLevel(0),
-	m_sourceId(0)
+	m_maxLevel(maxLevel),
+	m_followSymLinks(followSymLinks)
 {
 	if (m_dirName.empty() == false)
 	{
-		CrawlHistory history(PinotSettings::getInstance().m_historyDatabase);
+		CrawlHistory history(PinotSettings::getInstance().getHistoryDatabaseName());
 
 		if (isSource == true)
 		{
@@ -385,6 +440,28 @@ Signal3<void, const DocumentInfo&, const std::string&, bool>& DirectoryScannerTh
 	return m_signalFileFound;
 }
 
+void DirectoryScannerThread::cacheUpdate(const string &location, time_t mTime,
+	CrawlHistory &history)
+{
+	m_updateCache[location] = mTime;
+
+	if (m_updateCache.size() > 500)
+	{
+		flushUpdates(history);
+	}
+}
+
+void DirectoryScannerThread::flushUpdates(CrawlHistory &history)
+{
+	// Update these records
+	history.updateItems(m_updateCache, CrawlHistory::CRAWLED);
+	m_updateCache.clear();
+
+#ifdef DEBUG
+	cout << "DirectoryScannerThread::flushUpdates: flushed updates" << endl;
+#endif
+}
+
 void DirectoryScannerThread::foundFile(const DocumentInfo &docInfo)
 {
 	char labelStr[64];
@@ -401,23 +478,30 @@ void DirectoryScannerThread::foundFile(const DocumentInfo &docInfo)
 	m_signalFileFound(docInfo, labelStr, false);
 }
 
-bool DirectoryScannerThread::scanEntry(const string &entryName)
+bool DirectoryScannerThread::scanEntry(const string &entryName, CrawlHistory &history)
 {
-	CrawlHistory history(PinotSettings::getInstance().m_historyDatabase);
 	CrawlHistory::CrawlStatus status = CrawlHistory::UNKNOWN;
+	string location("file://" + entryName);
 	time_t itemDate;
 	struct stat fileStat;
 	int statSuccess = 0;
+	bool scanSuccess = true, reportFile = false, itemExists = false;
 
 	if (entryName.empty() == true)
 	{
+#ifdef DEBUG
+		cout << "DirectoryScannerThread::scanEntry: no name" << endl;
+#endif
 		return false;
 	}
 
 	// Skip . .. and dotfiles
-	Url urlObj("file://" + entryName);
+	Url urlObj(location);
 	if (urlObj.getFile()[0] == '.')
 	{
+#ifdef DEBUG
+		cout << "DirectoryScannerThread::scanEntry: skipped dotfile " << urlObj.getFile() << endl;
+#endif
 		return false;
 	}
 
@@ -431,66 +515,33 @@ bool DirectoryScannerThread::scanEntry(const string &entryName)
 		statSuccess = stat(entryName.c_str(), &fileStat);
 	}
 
+	// Is this item in the database already ?
+	itemExists = history.hasItem(location, status, itemDate);
+
 	if (statSuccess == -1)
 	{
-		return false;
+#ifdef DEBUG
+		cout << "DirectoryScannerThread::scanEntry: stat failed with error " << errno << " " << strerror(errno) << endl;
+#endif
+		scanSuccess = false;
 	}
-
-	bool itemExists = history.hasItem("file://" + entryName, status, itemDate);
-
 	// Is it a file or a directory ?
-	if (S_ISLNK(fileStat.st_mode))
+	else if (S_ISLNK(fileStat.st_mode))
 	{
 		// This won't happen when m_followSymLinks is true
+#ifdef DEBUG
+		cout << "DirectoryScannerThread::scanEntry: skipped symlink" << endl;
+#endif
 		return false;
 	}
 	else if (S_ISREG(fileStat.st_mode))
 	{
-		DocumentInfo docInfo;
-		bool reportFile = false;
-
-		docInfo.setLocation("file://" + entryName);
-
 		// Is this file blacklisted ?
 		// We have to check early so that if necessary the file's status stays at CRAWLING 
 		// and it is removed from the index at the end of this crawl
 		if (PinotSettings::getInstance().isBlackListed(entryName) == false)
 		{
-			if (itemExists == false)
-			{
-				// Record it
-				history.insertItem(docInfo.getLocation(), CrawlHistory::CRAWLED, m_sourceId, fileStat.st_mtime);
-#ifdef DEBUG
-				cout << "DirectoryScannerThread::scanEntry: reporting new file " << entryName << endl;
-#endif
-				reportFile = true;
-			}
-			else
-			{
-				// Update the record
-				history.updateItem(docInfo.getLocation(), CrawlHistory::CRAWLED, fileStat.st_mtime);
-
-				// Was it last crawled after it was modified ?
-				if (itemDate < fileStat.st_mtime)
-				{
-#ifdef DEBUG
-					cout << "DirectoryScannerThread::scanEntry: reporting modified file " << entryName << endl;
-#endif
-					// No, crawl and index it again
-					reportFile = true;
-				}
-			}
-		}
-
-		if (reportFile == true)
-		{
-			Url urlObj(docInfo.getLocation());
-
-			docInfo.setTitle(urlObj.getFile());
-			docInfo.setTimestamp(TimeConverter::toTimestamp(fileStat.st_mtime));
-			docInfo.setSize(fileStat.st_size);
-
-			foundFile(docInfo);
+			reportFile = true;
 		}
 	}
 	else if (S_ISDIR(fileStat.st_mode))
@@ -504,88 +555,177 @@ bool DirectoryScannerThread::scanEntry(const string &entryName)
 
 			// Open the directory
 			DIR *pDir = opendir(entryName.c_str());
-			if (pDir == NULL)
+			if (pDir != NULL)
 			{
-				return false;
-			}
 #ifdef DEBUG
-			cout << "DirectoryScannerThread::scanEntry: entering " << entryName << endl;
+				cout << "DirectoryScannerThread::scanEntry: entering " << entryName << endl;
 #endif
-
-			if (m_pMonitor != NULL)
-			{
-				// Monitor first so that we don't miss events
-				m_pMonitor->addLocation(entryName, true);
-			}
-
-			// Iterate through this directory's entries
-			struct dirent *pDirEntry = readdir(pDir);
-			while ((m_done == false) &&
-				(pDirEntry != NULL))
-			{
-				char *pEntryName = pDirEntry->d_name;
-
-				// Skip . .. and dotfiles
-				if ((pEntryName != NULL) &&
-					(pEntryName[0] != '.'))
+				if (m_pMonitor != NULL)
 				{
-					string subEntryName(entryName);
-
-					if (entryName[entryName.length() - 1] != '/')
-					{
-						subEntryName += "/";
-					}
-					subEntryName += pEntryName;
-
-					// Scan this entry
-					if (scanEntry(subEntryName) == false)
-					{
-#ifdef DEBUG
-						cout << "DirectoryScannerThread::scanEntry: failed to open "
-							<< subEntryName << endl;
-#endif
-					}
+					// Monitor first so that we don't miss events
+					m_pMonitor->addLocation(entryName, true);
 				}
 
-				// Next entry
-				pDirEntry = readdir(pDir);
-			}
+				// Iterate through this directory's entries
+				struct dirent *pDirEntry = readdir(pDir);
+				while ((m_done == false) &&
+					(pDirEntry != NULL))
+				{
+					char *pEntryName = pDirEntry->d_name;
+
+					// Skip . .. and dotfiles
+					if ((pEntryName != NULL) &&
+						(pEntryName[0] != '.'))
+					{
+						string subEntryName(entryName);
+
+						if (entryName[entryName.length() - 1] != '/')
+						{
+							subEntryName += "/";
+						}
+						subEntryName += pEntryName;
+
+						// Scan this entry
+						if (scanEntry(subEntryName, history) == false)
+						{
 #ifdef DEBUG
-			cout << "DirectoryScannerThread::scanEntry: done with " << entryName << endl;
+							cout << "DirectoryScannerThread::scanEntry: failed to open "
+								<< subEntryName << endl;
+#endif
+						}
+					}
+
+					// Next entry
+					pDirEntry = readdir(pDir);
+				}
+#ifdef DEBUG
+				cout << "DirectoryScannerThread::scanEntry: done with " << entryName << endl;
 #endif
 
-			// Close the directory
-			closedir(pDir);
-			--m_currentLevel;
+				// Close the directory
+				closedir(pDir);
+				--m_currentLevel;
+				reportFile = true;
+			}
+			else
+			{
+#ifdef DEBUG
+				cout << "DirectoryScannerThread::scanEntry: opendir failed with error " << errno << " " << strerror(errno) << endl;
+#endif
+				scanSuccess = false;
+			}
 		}
 	}
 	else
 	{
-		return false;
+#ifdef DEBUG
+		cout << "DirectoryScannerThread::scanEntry: unknown entry type" << endl;
+#endif
+		scanSuccess = false;
 	}
 
-	return true;
+	// Did an error occur ?
+	if (scanSuccess == false)
+	{
+		time_t timeNow = time(NULL);
+
+		// Record this error
+		if (itemExists == false)
+		{
+			history.insertItem(location, CrawlHistory::ERROR, m_sourceId, timeNow);
+		}
+		else
+		{
+			history.updateItem(location, CrawlHistory::ERROR, timeNow);
+		}
+	}
+	else if (reportFile == true)
+	{
+		if (itemExists == false)
+		{
+			// Record it
+			history.insertItem(location, CrawlHistory::CRAWLED, m_sourceId, fileStat.st_mtime);
+#ifdef DEBUG
+			cout << "DirectoryScannerThread::scanEntry: reporting new file " << entryName << endl;
+#endif
+		}
+		else
+		{
+			// Was it modified after the last crawl ?
+			if (itemDate >= fileStat.st_mtime)
+			{
+				// No, it wasn't
+				reportFile = false;
+			}
+
+			// History of modified files, especially the timestamp, is always updated
+			// Others' are updated only if we are doing a full scan because
+			// the status has to be reset to CRAWLED, so that they are not unindexed
+			if ((reportFile == true) ||
+				(m_fullScan == true))
+			{
+#ifdef DEBUG
+				cout << "DirectoryScannerThread::scanEntry: updating " << entryName << endl;
+#endif
+				cacheUpdate(location, fileStat.st_mtime, history);
+			}
+
+		}
+
+		if (reportFile == true)
+		{
+			DocumentInfo docInfo;
+
+			docInfo.setLocation(location);
+			docInfo.setTitle(urlObj.getFile());
+			if (S_ISDIR(fileStat.st_mode))
+			{
+				docInfo.setType("x-directory/normal");
+			}
+			else
+			{
+				// We might as well scan the file for its type now
+				docInfo.setType(MIMEScanner::scanFile(entryName));
+			}
+			docInfo.setTimestamp(TimeConverter::toTimestamp(fileStat.st_mtime));
+			docInfo.setSize(fileStat.st_size);
+
+			foundFile(docInfo);
+		}
+	}
+
+	return scanSuccess;
 }
 
 void DirectoryScannerThread::doWork(void)
 {
-	CrawlHistory history(PinotSettings::getInstance().m_historyDatabase);
+	CrawlHistory history(PinotSettings::getInstance().getHistoryDatabaseName());
 	set<string> deletedFiles;
+	Timer scanTimer;
 
 	if (m_dirName.empty() == true)
 	{
 		return;
 	}
+	scanTimer.start();
 
-	// Update this source's items status
-	history.updateItemsStatus(m_sourceId, CrawlHistory::CRAWLED, CrawlHistory::CRAWLING);
+	if (m_fullScan == true)
+	{
+		cout << "Doing a full scan on " << m_dirName << endl;
 
-	if (scanEntry(m_dirName) == false)
+		// Update this source's items status so that we can detect files that have been deleted
+		history.updateItemsStatus(m_sourceId, CrawlHistory::CRAWLED, CrawlHistory::CRAWLING);
+	}
+	// Remove errors
+	history.deleteItems(m_sourceId, CrawlHistory::ERROR);
+
+	if (scanEntry(m_dirName, history) == false)
 	{
 		m_status = _("Couldn't open directory");
 		m_status += " ";
 		m_status += m_dirName;
 	}
+	flushUpdates(history);
 
 	if (m_done == true)
 	{
@@ -595,28 +735,29 @@ void DirectoryScannerThread::doWork(void)
 		return;
 	}
 
-	// All files with status set to CRAWLING were not found in this crawl
-	// Chances are they were removed after the previous crawl
-	if ((m_pHandler != NULL) &&
-		(history.getSourceItems(m_sourceId, CrawlHistory::CRAWLING, deletedFiles) > 0))
+	if (m_fullScan == true)
 	{
-#ifdef DEBUG
-		cout << "DirectoryScannerThread::doWork: " << deletedFiles.size() << " files were deleted" << endl;
-#endif
-		for(set<string>::const_iterator fileIter = deletedFiles.begin();
-			fileIter != deletedFiles.end(); ++fileIter)
+		// All files with status set to CRAWLING were not found in this crawl
+		// Chances are they were removed after the last full scan
+		if ((m_pHandler != NULL) &&
+			(history.getSourceItems(m_sourceId, CrawlHistory::CRAWLING, deletedFiles) > 0))
 		{
-			// Inform the MonitorHandler
-			if (m_pHandler->fileDeleted(fileIter->substr(7)) == true)
+#ifdef DEBUG
+			cout << "DirectoryScannerThread::doWork: " << deletedFiles.size() << " files were deleted" << endl;
+#endif
+			for(set<string>::const_iterator fileIter = deletedFiles.begin();
+				fileIter != deletedFiles.end(); ++fileIter)
 			{
-				// Delete this item
-				history.deleteItem(*fileIter);
+				// Inform the MonitorHandler
+				if (m_pHandler->fileDeleted(fileIter->substr(7)) == true)
+				{
+					// Delete this item
+					history.deleteItem(*fileIter);
+				}
 			}
 		}
 	}
-#ifdef DEBUG
-	cout << "DirectoryScannerThread::doWork: done crawling " << m_dirName << endl;
-#endif
+	cout << "Scanned " << m_dirName << " in " << scanTimer.stop() << " ms" << endl;
 }
 
 DBusServletThread::DBusServletThread(DaemonState *pServer, DBusConnection *pConnection, DBusMessage *pRequest) :
@@ -673,11 +814,42 @@ bool DBusServletThread::mustQuit(void) const
 	return m_mustQuit;
 }
 
+bool DBusServletThread::runQuery(QueryProperties &queryProps, vector<string> &docIds)
+{
+	XapianEngine engine(PinotSettings::getInstance().m_daemonIndexLocation);
+
+	// Run the query
+	if (engine.runQuery(queryProps) == false)
+	{
+		return false;
+	}
+
+	docIds.clear();
+
+	const vector<DocumentInfo> &resultsList = engine.getResults();
+	for (vector<DocumentInfo>::const_iterator resultIter = resultsList.begin();
+		resultIter != resultsList.end(); ++resultIter)
+	{
+		unsigned int indexId = 0;
+		unsigned int docId = resultIter->getIsIndexed(indexId);
+
+		// We only need the document ID
+		if (docId > 0)
+		{
+			char docIdStr[64];
+
+			snprintf(docIdStr, 64, "%u", docId);
+			docIds.push_back(docIdStr);
+		}
+	}
+
+	return true;
+}
+
 void DBusServletThread::doWork(void)
 {
 	XapianIndex index(PinotSettings::getInstance().m_daemonIndexLocation);
 	DBusError error;
-	const char *pSender = dbus_message_get_sender(m_pRequest);
 	bool processedMessage = true, flushIndex = false;
 
 	if ((m_pServer == NULL) ||
@@ -690,6 +862,7 @@ void DBusServletThread::doWork(void)
 	dbus_error_init(&error);
 
 #ifdef DEBUG
+	const char *pSender = dbus_message_get_sender(m_pRequest);
 	if (pSender != NULL)
 	{
 		cout << "DBusServletThread::doWork: called by " << pSender << endl;
@@ -700,7 +873,52 @@ void DBusServletThread::doWork(void)
 	}
 #endif
 
-	if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "DeleteLabel") == TRUE)
+	if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "GetStatistics") == TRUE)
+	{
+		CrawlHistory history(PinotSettings::getInstance().getHistoryDatabaseName());
+		unsigned int crawledFilesCount = history.getItemsCount(CrawlHistory::CRAWLED);
+		unsigned int docsCount = index.getDocumentsCount();
+
+#ifdef DEBUG
+		cout << "DBusServletThread::doWork: received GetStatistics" << endl;
+#endif
+		// Prepare the reply
+		m_pReply = newDBusReply(m_pRequest);
+		if (m_pReply != NULL)
+		{
+			dbus_message_append_args(m_pReply,
+				DBUS_TYPE_UINT32, &crawledFilesCount,
+				DBUS_TYPE_UINT32, &docsCount,
+				DBUS_TYPE_INVALID);
+		}
+	}
+	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "RenameLabel") == TRUE)
+	{
+		char *pOldLabel = NULL;
+		char *pNewLabel = NULL;
+
+		if (dbus_message_get_args(m_pRequest, &error,
+			DBUS_TYPE_STRING, &pOldLabel,
+			DBUS_TYPE_STRING, &pNewLabel,
+			DBUS_TYPE_INVALID) == TRUE)
+		{
+#ifdef DEBUG
+			cout << "DBusServletThread::doWork: received RenameLabel " << pOldLabel << ", " << pNewLabel << endl;
+#endif
+			// Rename the label
+			flushIndex = index.renameLabel(pOldLabel, pNewLabel);
+
+			// Prepare the reply
+			m_pReply = newDBusReply(m_pRequest);
+			if (m_pReply != NULL)
+			{
+				dbus_message_append_args(m_pReply,
+					DBUS_TYPE_STRING, &pNewLabel,
+					DBUS_TYPE_INVALID);
+			}
+		}
+	}
+	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "DeleteLabel") == TRUE)
 	{
 		char *pLabel = NULL;
 
@@ -724,47 +942,6 @@ void DBusServletThread::doWork(void)
 			}
 		}
 	}
-	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "GetDocumentInfo") == TRUE)
-	{
-		unsigned int docId = 0;
-
-		if (dbus_message_get_args(m_pRequest, &error,
-			DBUS_TYPE_UINT32, &docId,
-			DBUS_TYPE_INVALID) == TRUE)
-		{
-			DocumentInfo docInfo;
-
-#ifdef DEBUG
-			cout << "DBusServletThread::doWork: received GetDocumentInfo " << docId << endl;
-#endif
-			if (index.getDocumentInfo(docId, docInfo) == true)
-			{
-				// Prepare the reply
-				m_pReply = newDBusReply(m_pRequest);
-				if (m_pReply != NULL)
-				{
-					string language(Languages::toEnglish(docInfo.getLanguage()));
-					const char *pTitle = docInfo.getTitle().c_str();
-					const char *pLocation = docInfo.getLocation().c_str();
-					const char *pType = docInfo.getType().c_str();
-					const char *pLanguage = language.c_str();
-
-					dbus_message_append_args(m_pReply,
-						DBUS_TYPE_STRING, &pTitle,
-						DBUS_TYPE_STRING, &pLocation,
-						DBUS_TYPE_STRING, &pType,
-						DBUS_TYPE_STRING, &pLanguage,
-						DBUS_TYPE_INVALID);
-				}
-			}
-			else
-			{
-				m_pReply = dbus_message_new_error(m_pRequest,
-					"de.berlios.Pinot.GetDocumentInfo",
-					"Unknown document");
-			}
-		}
-	}
 	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "GetDocumentLabels") == TRUE)
 	{
 		unsigned int docId = 0;
@@ -810,88 +987,6 @@ void DBusServletThread::doWork(void)
 			}
 		}
 	}
-	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "GetStatistics") == TRUE)
-	{
-		CrawlHistory history(PinotSettings::getInstance().m_historyDatabase);
-		unsigned int crawledFilesCount = history.getItemsCount();
-		unsigned int docsCount = index.getDocumentsCount();
-
-#ifdef DEBUG
-		cout << "DBusServletThread::doWork: received GetStatistics" << endl;
-#endif
-		// Prepare the reply
-		m_pReply = newDBusReply(m_pRequest);
-		if (m_pReply != NULL)
-		{
-			dbus_message_append_args(m_pReply,
-				DBUS_TYPE_UINT32, &crawledFilesCount,
-				DBUS_TYPE_UINT32, &docsCount,
-				DBUS_TYPE_INVALID);
-		}
-	}
-	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "RenameLabel") == TRUE)
-	{
-		char *pOldLabel = NULL;
-		char *pNewLabel = NULL;
-
-		if (dbus_message_get_args(m_pRequest, &error,
-			DBUS_TYPE_STRING, &pOldLabel,
-			DBUS_TYPE_STRING, &pNewLabel,
-			DBUS_TYPE_INVALID) == TRUE)
-		{
-#ifdef DEBUG
-			cout << "DBusServletThread::doWork: received RenameLabel " << pOldLabel << ", " << pNewLabel << endl;
-#endif
-			// Rename the label
-			flushIndex = index.renameLabel(pOldLabel, pNewLabel);
-
-			// Prepare the reply
-			m_pReply = newDBusReply(m_pRequest);
-			if (m_pReply != NULL)
-			{
-				dbus_message_append_args(m_pReply,
-					DBUS_TYPE_STRING, &pNewLabel,
-					DBUS_TYPE_INVALID);
-			}
-		}
-	}
-	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "SetDocumentInfo") == TRUE)
-	{
-		char *pTitle = NULL;
-		char *pLocation = NULL;
-		char *pType = NULL;
-		char *pLanguage = NULL;
-		unsigned int docId = 0;
-
-		if (dbus_message_get_args(m_pRequest, &error,
-			DBUS_TYPE_UINT32, &docId,
-			DBUS_TYPE_STRING, &pTitle,
-			DBUS_TYPE_STRING, &pLocation,
-			DBUS_TYPE_STRING, &pType,
-			DBUS_TYPE_STRING, &pLanguage,
-			DBUS_TYPE_INVALID) == TRUE)
-		{
-			DocumentInfo docInfo(pTitle, pLocation, pType,
-				((pLanguage != NULL) ? Languages::toLocale(pLanguage) : ""));
-
-#ifdef DEBUG
-			cout << "DBusServletThread::doWork: received SetDocumentInfo " << docId << ", " << pTitle
-				<< ", " << pLocation << ", " << pType << ", " << pLanguage << endl;
-#endif
-
-			// Update the document info
-			flushIndex = index.updateDocumentInfo(docId, docInfo);
-
-			// Prepare the reply
-			m_pReply = newDBusReply(m_pRequest);
-			if (m_pReply != NULL)
-			{
-				dbus_message_append_args(m_pReply,
-					DBUS_TYPE_UINT32, &docId,
-					DBUS_TYPE_INVALID);
-			}
-		}
-	}
 	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "SetDocumentLabels") == TRUE)
 	{
 		char **ppLabels = NULL;
@@ -969,7 +1064,7 @@ void DBusServletThread::doWork(void)
 				labels.insert(ppLabels[labelIndex]);
 			}
 #ifdef DEBUG
-			cout << "DBusServletThread::doWork: received SetDocumentLabels on " << docIds.size()
+			cout << "DBusServletThread::doWork: received SetDocumentsLabels on " << docIds.size()
 				<< " IDs, " << labelsCount << " labels" << ", " << resetLabels << endl;
 #endif
 			// Set labels
@@ -989,6 +1084,77 @@ void DBusServletThread::doWork(void)
 			}
 		}
 	}
+	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "GetDocumentInfo") == TRUE)
+	{
+		unsigned int docId = 0;
+
+		if (dbus_message_get_args(m_pRequest, &error,
+			DBUS_TYPE_UINT32, &docId,
+			DBUS_TYPE_INVALID) == TRUE)
+		{
+			DocumentInfo docInfo;
+
+#ifdef DEBUG
+			cout << "DBusServletThread::doWork: received GetDocumentInfo on " << docId << endl;
+#endif
+			if (index.getDocumentInfo(docId, docInfo) == true)
+			{
+				// Prepare the reply
+				m_pReply = newDBusReply(m_pRequest);
+				if (m_pReply != NULL)
+				{
+					DBusMessageIter iter;
+
+					dbus_message_iter_init_append(m_pReply, &iter);
+					if (DBusXapianIndex::documentInfoToDBus(&iter, 0, docInfo) == false)
+					{
+						dbus_message_unref(m_pReply);
+						m_pReply = dbus_message_new_error(m_pRequest,
+							"de.berlios.Pinot.GetDocumentInfo",
+							"Unknown error");
+					}
+				}
+			}
+			else
+			{
+				m_pReply = dbus_message_new_error(m_pRequest,
+					"de.berlios.Pinot.GetDocumentInfo",
+					"Unknown document");
+			}
+		}
+	}
+	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "SetDocumentInfo") == TRUE)
+	{
+		DBusMessageIter iter;
+		DocumentInfo docInfo;
+		unsigned int docId = 0;
+
+		dbus_message_iter_init(m_pRequest, &iter);
+		if (DBusXapianIndex::documentInfoFromDBus(&iter, docId, docInfo) == false)
+		{
+			m_pReply = dbus_message_new_error(m_pRequest,
+				"de.berlios.Pinot.SetDocumentInfo",
+				"Unknown error");
+		}
+		else
+		{
+#ifdef DEBUG
+			cout << "DBusServletThread::doWork: received SetDocumentInfo on " << docId << endl;
+#endif
+
+			// Update the document info
+			flushIndex = index.updateDocumentInfo(docId, docInfo);
+
+			// Prepare the reply
+			m_pReply = newDBusReply(m_pRequest);
+			if (m_pReply != NULL)
+			{
+				dbus_message_append_args(m_pReply,
+					DBUS_TYPE_UINT32, &docId,
+					DBUS_TYPE_INVALID);
+			}
+		}
+	}
 	else if (dbus_message_is_method_call(m_pRequest, "de.berlios.Pinot", "SimpleQuery") == TRUE)
 	{
 		char *pSearchText = NULL;
@@ -999,7 +1165,6 @@ void DBusServletThread::doWork(void)
 			DBUS_TYPE_UINT32, &maxHits,
 			DBUS_TYPE_INVALID) == TRUE)
 		{
-			XapianEngine engine(PinotSettings::getInstance().m_daemonIndexLocation);
 			bool replyWithError = true;
 
 #ifdef DEBUG
@@ -1008,33 +1173,20 @@ void DBusServletThread::doWork(void)
 			if (pSearchText != NULL)
 			{
 				QueryProperties queryProps("DBUS", pSearchText);
+				vector<string> docIds;
 
 				// Run the query
-				engine.setMaxResultsCount(maxHits);
-				if (engine.runQuery(queryProps) == true)
+				queryProps.setMaximumResultsCount(maxHits);
+				if (runQuery(queryProps, docIds) == true)
 				{
-					const vector<Result> &resultsList = engine.getResults();
-					vector<string> docIds;
 					m_pArray = g_ptr_array_new();
 
-					for (vector<Result>::const_iterator resultIter = resultsList.begin();
-						resultIter != resultsList.end(); ++resultIter)
-					{
-						// We only need the document ID
-						unsigned int docId = index.hasDocument(resultIter->getLocation());
-						if (docId > 0)
-						{
-							char docIdStr[64];
-							snprintf(docIdStr, 64, "%u", docId);
-							docIds.push_back(docIdStr);
-						}
-					}
-
 					for (vector<string>::const_iterator docIter = docIds.begin();
 						docIter != docIds.end(); ++docIter)
 					{
 #ifdef DEBUG
-						cout << "DBusServletThread::doWork: adding result " << m_pArray->len << " " << *docIter << endl;
+						cout << "DBusServletThread::doWork: adding result "
+							<< m_pArray->len << " " << *docIter << endl;
 #endif
 						g_ptr_array_add(m_pArray, const_cast<char*>(docIter->c_str()));
 					}
@@ -1111,6 +1263,25 @@ void DBusServletThread::doWork(void)
 			}
 		}
 	}
+	else if (dbus_message_is_method_call(m_pRequest, "org.freedesktop.DBus.Introspectable", "Introspect") == TRUE)
+	{
+#ifdef DEBUG
+		cout << "DBusServletThread::doWork: received Introspect" << endl;
+#endif
+		if (loadXMLDescription() == true)
+		{
+			// Prepare the reply
+			m_pReply = newDBusReply(m_pRequest);
+			if (m_pReply != NULL)
+			{
+				const char *pXmlData = g_xmlDescription.c_str();
+
+				dbus_message_append_args(m_pReply,
+					DBUS_TYPE_STRING, &pXmlData,
+					DBUS_TYPE_INVALID);
+			}
+		}
+	}
 	else
 	{
 #ifdef DEBUG
