@@ -18,6 +18,7 @@
 
 #include <unistd.h>
 #include <stdlib.h>
+#include <stdio.h>
 #include <fcntl.h>
 #include <string.h>
 #include <signal.h>
@@ -30,20 +31,21 @@
 #include <sigc++/slot.h>
 #include <glibmm/miscutils.h>
 
-#include "HtmlTokenizer.h"
-#include "XmlTokenizer.h"
 #include "MIMEScanner.h"
-#include "TokenizerFactory.h"
 #include "StringManip.h"
 #include "TimeConverter.h"
 #include "Url.h"
+#include "HtmlFilter.h"
+#include "FilterFactory.h"
+#include "FilterUtils.h"
+#include "FilterWrapper.h"
+#include "DBusXapianIndex.h"
 #include "XapianDatabase.h"
 #include "ActionQueue.h"
+#include "CrawlHistory.h"
 #include "QueryHistory.h"
-#include "IndexedDocument.h"
 #include "DownloaderFactory.h"
 #include "SearchEngineFactory.h"
-#include "IndexFactory.h"
 #include "config.h"
 #include "NLS.h"
 #include "PinotSettings.h"
@@ -67,22 +69,6 @@ public:
 	}
 };
 
-// A function object to delete threads with for_each()
-struct DeleteThreadFunc
-{
-public:
-	void operator()(map<WorkerThread *, Thread *>::value_type &p)
-	{
-#ifdef DEBUG
-		cout << "DeleteThreadFunc: waiting for thread " << p.first->getId() << endl;
-#endif
-		// FIXME: the documentation says resources of the thread, including the Thread object
-		// are released by join()
-		p.second->join();
-		delete p.first;
-	}
-};
-
 Dispatcher WorkerThread::m_dispatcher;
 pthread_mutex_t WorkerThread::m_dispatcherMutex = PTHREAD_MUTEX_INITIALIZER;
 bool WorkerThread::m_immediateFlush = true;
@@ -98,7 +84,6 @@ void WorkerThread::immediateFlush(bool doFlush)
 }
 
 WorkerThread::WorkerThread() :
-	m_joinable(true),
 	m_id(0),
 	m_background(false),
 	m_done(false)
@@ -139,7 +124,8 @@ Glib::Thread *WorkerThread::start(void)
 #ifdef DEBUG
 	cout << "WorkerThread::start: " << getType() << " " << m_id << endl;
 #endif
-	return Thread::create(slot_class(*this, &WorkerThread::threadHandler), m_joinable);
+	// Create non-joinable threads
+	return Thread::create(slot_class(*this, &WorkerThread::threadHandler), false);
 }
 
 bool WorkerThread::isDone(void) const
@@ -183,9 +169,6 @@ void WorkerThread::emitSignal(void)
 		cout << "WorkerThread::emitSignal: signaling end of thread " << m_id << endl;
 #endif
 		m_dispatcher();
-#ifdef DEBUG
-		cout << "WorkerThread::emitSignal: signaled end of thread " << m_id << endl;
-#endif
 
 		pthread_mutex_unlock(&m_dispatcherMutex);
 	}
@@ -277,7 +260,6 @@ WorkerThread *ThreadsManager::get_thread(void)
 			{
 				// This one will do...
 				pWorkerThread = threadIter->first;
-				threadIter->second->join();
 				// Remove it
 				m_threads.erase(threadIter);
 				break;
@@ -353,28 +335,7 @@ ustring ThreadsManager::index_document(const DocumentInfo &docInfo)
 		return status;
 	}
 
-	// Is it an update ?
-	IndexInterface *pIndex = PinotSettings::getInstance().getIndex(m_defaultIndexLocation);
-	if (pIndex == NULL)
-	{
-		ustring status = _("Index error on");
-		status += " ";
-		status += m_defaultIndexLocation;
-		return status;
-	}
-
-	unsigned int docId = pIndex->hasDocument(docInfo.getLocation());
-	if (docId > 0)
-	{
-		// Yes, it is
-		start_thread(new IndexingThread(docInfo, docId, m_defaultIndexLocation));
-	}
-	else
-	{
-		// This is a new document
-		start_thread(new IndexingThread(docInfo, docId, m_defaultIndexLocation));
-	}
-	delete pIndex;
+	start_thread(new IndexingThread(docInfo, m_defaultIndexLocation));
 
 	return "";
 }
@@ -451,8 +412,6 @@ void ThreadsManager::stop_threads(void)
 		{
 			// Stop threads
 			for_each(m_threads.begin(), m_threads.end(), StopThreadFunc());
-			// Join them
-			for_each(m_threads.begin(), m_threads.end(), DeleteThreadFunc());
 			m_threads.clear();
 
 			unlock_threads();
@@ -533,7 +492,7 @@ ustring ThreadsManager::queue_index(const DocumentInfo &docInfo)
 
 	if (addToQueue == true)
 	{
-		ActionQueue queue(PinotSettings::getInstance().m_historyDatabase, get_application_name());
+		ActionQueue queue(PinotSettings::getInstance().getHistoryDatabaseName(), get_application_name());
 
 		queue.pushItem(ActionQueue::INDEX, docInfo);
 
@@ -576,23 +535,39 @@ bool ThreadsManager::pop_queue(const string &urlWasIndexed)
 		// Get an item ?
 		if (getItem == true)
 		{
-			ActionQueue queue(PinotSettings::getInstance().m_historyDatabase, get_application_name());
+			ActionQueue queue(PinotSettings::getInstance().getHistoryDatabaseName(), get_application_name());
 			ActionQueue::ActionType type;
 			DocumentInfo docInfo;
+			string previousLocation;
 
 			while (queue.popItem(type, docInfo) == true)
 			{
+				ustring status;
+
 				if (type != ActionQueue::INDEX)
 				{
 					continue;
 				}
 
-				ustring status = index_document(docInfo);
+				if (docInfo.getLocation() == previousLocation)
+				{
+					// Something dodgy is going on, we got the same item twice !
+					status = previousLocation;
+					status += " ";
+					status += _("is already being indexed");
+				}
+				else
+				{
+					status = index_document(docInfo);
+				}
+
 				if (status.empty() == true)
 				{
 					foundItem = true;
 					break;
 				}
+
+				previousLocation = docInfo.getLocation();
 			}
 		}
 	}
@@ -612,14 +587,47 @@ void ThreadsManager::get_statistics(unsigned int &queueSize)
 	}
 }
 
-IndexBrowserThread::IndexBrowserThread(const string &indexName,
-	const string &labelName, unsigned int maxDocsCount, unsigned int startDoc) :
+ListerThread::ListerThread(const string &indexName, unsigned int startDoc) :
 	WorkerThread(),
 	m_indexName(indexName),
-	m_labelName(labelName),
-	m_indexDocsCount(0),
-	m_maxDocsCount(maxDocsCount),
-	m_startDoc(startDoc)
+	m_startDoc(startDoc),
+	m_documentsCount(0)
+{
+}
+
+ListerThread::~ListerThread()
+{
+}
+
+string ListerThread::getType(void) const
+{
+	return "ListerThread";
+}
+
+string ListerThread::getIndexName(void) const
+{
+	return m_indexName;
+}
+
+unsigned int ListerThread::getStartDoc(void) const
+{
+	return m_startDoc;
+}
+
+const vector<DocumentInfo> &ListerThread::getDocuments(void) const
+{
+	return m_documentsList;
+}
+
+unsigned int ListerThread::getDocumentsCount(void) const
+{
+	return m_documentsCount;
+}
+
+IndexBrowserThread::IndexBrowserThread(const string &indexName,
+	unsigned int maxDocsCount, unsigned int startDoc) :
+	ListerThread(indexName, startDoc),
+	m_maxDocsCount(maxDocsCount)
 {
 }
 
@@ -627,31 +635,6 @@ IndexBrowserThread::~IndexBrowserThread()
 {
 }
 
-string IndexBrowserThread::getType(void) const
-{
-	return "IndexBrowserThread";
-}
-
-string IndexBrowserThread::getIndexName(void) const
-{
-	return m_indexName;
-}
-
-string IndexBrowserThread::getLabelName(void) const
-{
-	return m_labelName;
-}
-
-unsigned int IndexBrowserThread::getDocumentsCount(void) const
-{
-	return m_indexDocsCount;
-}
-
-const vector<IndexedDocument> &IndexBrowserThread::getDocuments(void) const
-{
-	return m_documentsList;
-}
-
 bool IndexBrowserThread::stop(void)
 {
 	m_done = true;
@@ -691,8 +674,8 @@ void IndexBrowserThread::doWork(void)
 		return;
 	}
 
-	m_indexDocsCount = pIndex->getDocumentsCount(m_labelName);
-	if (m_indexDocsCount == 0)
+	m_documentsCount = pIndex->getDocumentsCount();
+	if (m_documentsCount == 0)
 	{
 #ifdef DEBUG
 		cout << "IndexBrowserThread::doWork: no documents" << endl;
@@ -701,19 +684,14 @@ void IndexBrowserThread::doWork(void)
 	}
 
 #ifdef DEBUG
-	cout << "IndexBrowserThread::doWork: " << m_maxDocsCount << " off " << m_indexDocsCount
-		<< " documents to browse, starting at " << m_startDoc << endl;
+	cout << "IndexBrowserThread::doWork: " << m_maxDocsCount << " off " << m_documentsCount
+		<< " documents to browse, starting at position " << m_startDoc << endl;
 #endif
-	if (m_labelName.empty() == true)
-	{
-		pIndex->listDocuments(docIDList, m_maxDocsCount, m_startDoc);
-	}
-	else
-	{
-		pIndex->listDocumentsWithLabel(m_labelName, docIDList, m_maxDocsCount, m_startDoc);
-	}
+	pIndex->listDocuments(docIDList, m_maxDocsCount, m_startDoc);
 
+	m_documentsList.clear();
 	m_documentsList.reserve(m_maxDocsCount);
+
 	for (set<unsigned int>::iterator iter = docIDList.begin(); iter != docIDList.end(); ++iter)
 	{
 		if (m_done == true)
@@ -723,8 +701,6 @@ void IndexBrowserThread::doWork(void)
 
 		// Get the document ID
 		unsigned int docId = (*iter);
-		// ...and the document URL
-		string url = XapianDatabase::buildUrl(mapIter->second, docId);
 
 		DocumentInfo docInfo;
 		if (pIndex->getDocumentInfo(docId, docInfo) == true)
@@ -735,10 +711,11 @@ void IndexBrowserThread::doWork(void)
 				type = "text/html";
 			}
 
-			IndexedDocument indexedDoc(docInfo.getTitle(), url, docInfo.getLocation(),
+			DocumentInfo indexedDoc(docInfo.getTitle(), docInfo.getLocation(),
 				type, docInfo.getLanguage());
 			indexedDoc.setTimestamp(docInfo.getTimestamp());
 			indexedDoc.setSize(docInfo.getSize());
+			indexedDoc.setIsIndexed(PinotSettings::getInstance().getIndexId(m_indexName), docId);
 
 			// Insert that document
 			m_documentsList.push_back(indexedDoc);
@@ -752,13 +729,19 @@ void IndexBrowserThread::doWork(void)
 }
 
 QueryingThread::QueryingThread(const string &engineName, const string &engineDisplayableName,
-	const string &engineOption, const QueryProperties &queryProps) :
-	WorkerThread(),
+	const string &engineOption, const QueryProperties &queryProps,
+	unsigned int startDoc, bool listingIndex) :
+	ListerThread(engineDisplayableName, startDoc),
 	m_engineName(engineName),
 	m_engineDisplayableName(engineDisplayableName),
 	m_engineOption(engineOption),
-	m_queryProps(queryProps)
+	m_queryProps(queryProps),
+	m_listingIndex(listingIndex)
 {
+#ifdef DEBUG
+	cout << "QueryingThread::QueryingThread: engine " << m_engineName << ", " << engineOption
+		<< ", mode " << m_listingIndex << endl;
+#endif
 }
 
 QueryingThread::~QueryingThread()
@@ -767,6 +750,11 @@ QueryingThread::~QueryingThread()
 
 string QueryingThread::getType(void) const
 {
+	if (m_listingIndex == true)
+	{
+		return ListerThread::getType();
+	}
+
 	return "QueryingThread";
 }
 
@@ -780,14 +768,9 @@ QueryProperties QueryingThread::getQuery(void) const
 	return m_queryProps;
 }
 
-const vector<Result> &QueryingThread::getResults(string &charset) const
+string QueryingThread::getCharset(void) const
 {
-	charset = m_resultsCharset;
-#ifdef DEBUG
-	cout << "QueryingThread::getResults: charset for " << m_engineDisplayableName << " is " << charset << endl;
-#endif
-
-	return m_resultsList;
+	return m_resultsCharset;
 }
 
 bool QueryingThread::stop(void)
@@ -796,8 +779,140 @@ bool QueryingThread::stop(void)
 	return true;
 }
 
+void QueryingThread::processResults(const vector<DocumentInfo> &resultsList)
+{
+	PinotSettings &settings = PinotSettings::getInstance();
+	IndexInterface *pDocsIndex = NULL;
+	IndexInterface *pDaemonIndex = NULL;
+	unsigned int indexId = 0;
+	bool isIndexQuery = false;
+
+	// Are we querying an index ?
+	if (m_engineName == "xapian")
+	{
+		// Internal index ?
+		if (m_engineOption == settings.m_docsIndexLocation)
+		{
+			indexId = settings.getIndexId(_("My Web Pages"));
+			isIndexQuery = true;
+		}
+		else if (m_engineOption == settings.m_daemonIndexLocation)
+		{
+			indexId = settings.getIndexId(_("My Documents"));
+			isIndexQuery = true;
+		}
+	}
+
+	// Will we have to query internal indices ?
+	if (isIndexQuery == false)
+	{
+		pDocsIndex = settings.getIndex(settings.m_docsIndexLocation);
+		pDaemonIndex = settings.getIndex(settings.m_daemonIndexLocation);
+	}
+
+	// Copy the results list
+	for (vector<DocumentInfo>::const_iterator resultIter = resultsList.begin();
+		resultIter != resultsList.end(); ++resultIter)
+	{
+		DocumentInfo current(*resultIter);
+		string title(_("No title"));
+		string location(current.getLocation());
+		string language(current.getLanguage());
+		unsigned int docId = 0;
+
+		// The title may contain formatting
+		if (current.getTitle().empty() == false)
+		{
+			title = FilterUtils::stripMarkup(current.getTitle());
+		}
+		current.setTitle(title);
+#ifdef DEBUG
+		cout << "QueryingThread::processResults: title is " << title << endl;
+#endif
+
+		// Use the query's language if the result's is unknown
+		if (language.empty() == true)
+		{
+			language = m_queryProps.getLanguage();
+		}
+		current.setLanguage(language);
+
+		if (isIndexQuery == true)
+		{
+			unsigned int tmpId = 0;
+
+			// The index engine should have set this
+			docId = current.getIsIndexed(tmpId);
+		}
+
+		// Is this in one of the indexes ?
+		if ((pDocsIndex != NULL) &&
+			(pDocsIndex->isGood() == true))
+		{
+			docId = pDocsIndex->hasDocument(location);
+			if (docId > 0)
+			{
+				indexId = settings.getIndexId(_("My Web Pages"));
+			}
+		}
+		if ((pDaemonIndex != NULL) &&
+			(pDaemonIndex->isGood() == true) &&
+			(docId == 0))
+		{
+			docId = pDaemonIndex->hasDocument(location);
+			if (docId > 0)
+			{
+				indexId = settings.getIndexId(_("My Documents"));
+			}
+		}
+
+		if (docId > 0)
+		{
+			current.setIsIndexed(indexId, docId);
+#ifdef DEBUG
+			cout << "QueryingThread::processResults: found in index " << indexId << endl;
+#endif
+		}
+#ifdef DEBUG
+		else cout << "QueryingThread::processResults: not found in any index" << endl;
+#endif
+
+		m_documentsList.push_back(current);
+	}
+
+	if (pDocsIndex != NULL)
+	{
+		delete pDocsIndex;
+	}
+	if (pDaemonIndex != NULL)
+	{
+		delete pDaemonIndex;
+	}
+}
+
+void QueryingThread::processResults(const vector<DocumentInfo> &resultsList,
+	unsigned int indexId)
+{
+	unsigned int zeroId = 0;
+
+	// Copy the results list
+	for (vector<DocumentInfo>::const_iterator resultIter = resultsList.begin();
+		resultIter != resultsList.end(); ++resultIter)
+	{
+		DocumentInfo current(*resultIter);
+
+		// The engine has no notion of index IDs
+		unsigned int docId = current.getIsIndexed(zeroId);
+		current.setIsIndexed(indexId, docId);
+
+		m_documentsList.push_back(current);
+	}
+}
+
 void QueryingThread::doWork(void)
 {
+	PinotSettings &settings = PinotSettings::getInstance();
+
 	// Get the SearchEngine
 	SearchEngineInterface *pEngine = SearchEngineFactory::getSearchEngine(m_engineName, m_engineOption);
 	if (pEngine == NULL)
@@ -808,11 +923,22 @@ void QueryingThread::doWork(void)
 		return;
 	}
 
-	// Set the maximum number of results
-	pEngine->setMaxResultsCount(m_queryProps.getMaximumResultsCount());
+	// Set up the proxy
+	DownloaderInterface *pDownloader = pEngine->getDownloader();
+	if ((pDownloader != NULL) &&
+		(settings.m_proxyEnabled == true) &&
+		(settings.m_proxyAddress.empty() == false))
+	{
+		char portStr[64];
+
+		pDownloader->setSetting("proxyaddress", settings.m_proxyAddress);
+		snprintf(portStr, 64, "%u", settings.m_proxyPort);
+		pDownloader->setSetting("proxyport", portStr);
+		pDownloader->setSetting("proxytype", settings.m_proxyType);
+	}
 
 	// Run the query
-	if (pEngine->runQuery(m_queryProps) == false)
+	if (pEngine->runQuery(m_queryProps, m_startDoc) == false)
 	{
 		m_status = _("Couldn't run query on search engine");
 		m_status += " ";
@@ -820,51 +946,34 @@ void QueryingThread::doWork(void)
 	}
 	else
 	{
-		const vector<Result> &resultsList = pEngine->getResults();
+		const vector<DocumentInfo> &resultsList = pEngine->getResults();
 
-		m_resultsList.clear();
-		m_resultsList.reserve(resultsList.size());
-		m_resultsCharset = pEngine->getResultsCharset();
-
-		// Copy the results list
-		for (vector<Result>::const_iterator resultIter = resultsList.begin();
-			resultIter != resultsList.end(); ++resultIter)
-		{
-			string title(_("No title"));
-			string language(resultIter->getLanguage());
-
-			// The title may contain formatting
-			if (resultIter->getTitle().empty() == false)
-			{
-				title = XmlTokenizer::stripTags(resultIter->getTitle());
-			}
+		m_documentsList.clear();
+		m_documentsList.reserve(resultsList.size());
+		m_documentsCount = pEngine->getResultsCountEstimate();
 #ifdef DEBUG
-			cout << "QueryingThread::doWork: title is " << title << endl;
+		cout << "QueryingThread::doWork: " << resultsList.size() << " off " << m_documentsCount
+			<< " results to process, starting at position " << m_startDoc << endl;
 #endif
 
-			// Use the query's language if the result's is unknown
-			if (language.empty() == true)
-			{
-				language = m_queryProps.getLanguage();
-			}
-
-			m_resultsList.push_back(Result(resultIter->getLocation(),
-				title,
-				resultIter->getExtract(),
-				language,
-				resultIter->getScore()));
+		m_resultsCharset = pEngine->getResultsCharset();
+		if (m_listingIndex == false)
+		{
+			processResults(resultsList);
+		}
+		else
+		{
+			processResults(resultsList,
+				PinotSettings::getInstance().getIndexId(m_engineDisplayableName));
 		}
 	}
 
 	delete pEngine;
 }
 
-ExpandQueryThread::ExpandQueryThread(const string &engineName,
-	const string &engineOption, const QueryProperties &queryProps,
-	const set<unsigned int> &relevantDocs) :
+ExpandQueryThread::ExpandQueryThread(const QueryProperties &queryProps,
+	const set<string> &relevantDocs) :
 	WorkerThread(),
-	m_engineName(engineName),
-	m_engineOption(engineOption),
 	m_queryProps(queryProps)
 {
 	copy(relevantDocs.begin(), relevantDocs.end(),
@@ -898,8 +1007,31 @@ bool ExpandQueryThread::stop(void)
 
 void ExpandQueryThread::doWork(void)
 {
+	IndexInterface *pIndex = PinotSettings::getInstance().getIndex("MERGED");
+	set<unsigned int> relevantDocIds;
+
+	if ((pIndex == NULL) ||
+		(pIndex->isGood() == false))
+	{
+		m_status = _("Index error on");
+		m_status += " MERGED";
+		if (pIndex != NULL)
+		{
+			delete pIndex;
+		}
+		return;
+	}
+
+	for (set<string>::iterator locationIter = m_relevantDocs.begin();
+		locationIter != m_relevantDocs.end(); ++locationIter)
+	{
+		relevantDocIds.insert(pIndex->hasDocument(*locationIter));
+	}
+
+	delete pIndex;
+
 	// Get the SearchEngine
-	SearchEngineInterface *pEngine = SearchEngineFactory::getSearchEngine(m_engineName, m_engineOption);
+	SearchEngineInterface *pEngine = SearchEngineFactory::getSearchEngine("xapian", "MERGED");
 	if (pEngine == NULL)
 	{
 		m_status = _("Couldn't create search engine");
@@ -908,17 +1040,13 @@ void ExpandQueryThread::doWork(void)
 		return;
 	}
 
-	// Set the maximum number of results
-	pEngine->setMaxResultsCount(m_queryProps.getMaximumResultsCount());
 	// Set whether to expand the query
-	pEngine->setQueryExpansion(m_relevantDocs);
+	pEngine->setQueryExpansion(relevantDocIds);
 
 	// Run the query
 	if (pEngine->runQuery(m_queryProps) == false)
 	{
 		m_status = _("Couldn't run query on search engine");
-		m_status += " ";
-		m_status += m_engineName;
 	}
 	else
 	{
@@ -932,12 +1060,22 @@ void ExpandQueryThread::doWork(void)
 }
 
 LabelUpdateThread::LabelUpdateThread(const set<string> &labelsToDelete,
-	const map<string, string> &labelsToRename)
+	const map<string, string> &labelsToRename) :
+	WorkerThread()
 {
 	copy(labelsToDelete.begin(), labelsToDelete.end(), inserter(m_labelsToDelete, m_labelsToDelete.begin()));
 	copy(labelsToRename.begin(), labelsToRename.end(), inserter(m_labelsToRename, m_labelsToRename.begin()));
 }
 
+LabelUpdateThread::LabelUpdateThread(const string &labelName,
+	const set<unsigned int> &docsIds, const set<unsigned int> &daemonIds) :
+	WorkerThread(),
+	m_labelName(labelName)
+{
+	copy(docsIds.begin(), docsIds.end(), inserter(m_docsIds, m_docsIds.begin()));
+	copy(daemonIds.begin(), daemonIds.end(), inserter(m_daemonIds, m_daemonIds.begin()));
+}
+
 LabelUpdateThread::~LabelUpdateThread()
 {
 }
@@ -974,6 +1112,21 @@ void LabelUpdateThread::doWork(void)
 		return;
 	}
 
+	// Apply the new label to existing documents
+	if (m_labelName.empty() == false)
+	{
+		set<string> labels;
+
+		labels.insert(m_labelName);
+		if (m_docsIds.empty() == false)
+		{
+			pDocsIndex->setDocumentsLabels(m_docsIds, labels, false);
+		}
+		if (m_daemonIds.empty() == false)
+		{
+			pDaemonIndex->setDocumentsLabels(m_daemonIds, labels, false);
+		}
+	}
 	// Delete labels
 	for (set<string>::iterator iter = m_labelsToDelete.begin(); iter != m_labelsToDelete.end(); ++iter)
 	{
@@ -1052,6 +1205,20 @@ void DownloadingThread::doWork(void)
 	}
 	else if (m_done == false)
 	{
+		PinotSettings &settings = PinotSettings::getInstance();
+
+		// Set up the proxy
+		if ((settings.m_proxyEnabled == true) &&
+			(settings.m_proxyAddress.empty() == false))
+		{
+			char portStr[64];
+
+			m_pDownloader->setSetting("proxyaddress", settings.m_proxyAddress);
+			snprintf(portStr, 64, "%u", settings.m_proxyPort);
+			m_pDownloader->setSetting("proxyport", portStr);
+			m_pDownloader->setSetting("proxytype", settings.m_proxyType);
+		}
+
 		m_pDoc = m_pDownloader->retrieveUrl(m_docInfo);
 	}
 
@@ -1063,26 +1230,15 @@ void DownloadingThread::doWork(void)
 	}
 }
 
-IndexingThread::IndexingThread(const DocumentInfo &docInfo, unsigned int docId,
-	const string &indexLocation, bool allowAllMIMETypes) :
+IndexingThread::IndexingThread(const DocumentInfo &docInfo, const string &indexLocation,
+	bool allowAllMIMETypes) :
 	DownloadingThread(docInfo),
 	m_docInfo(docInfo),
-	m_docId(docId),
+	m_docId(0),
 	m_indexLocation(indexLocation),
-	m_allowAllMIMETypes(allowAllMIMETypes)
+	m_allowAllMIMETypes(allowAllMIMETypes),
+	m_update(false)
 {
-	if (m_docId > 0)
-	{
-		// Ignore robots directives on updates
-		m_ignoreRobotsDirectives = true;
-		m_update = true;
-	}
-	else
-	{
-		m_ignoreRobotsDirectives = PinotSettings::getInstance().m_ignoreRobotsDirectives;
-		// This is not an update
-		m_update = false;
-	}
 }
 
 IndexingThread::~IndexingThread()
@@ -1129,7 +1285,6 @@ void IndexingThread::doWork(void)
 {
 	IndexInterface *pIndex = PinotSettings::getInstance().getIndex(m_indexLocation);
 	Url thisUrl(m_docInfo.getLocation());
-	Tokenizer::DataNeeds dataNeeds;
 	bool doDownload = true;
 
 	// First things first, get the index
@@ -1146,14 +1301,22 @@ void IndexingThread::doWork(void)
 		return;
 	}
 
+	// Is it an update ?
+	m_docId = pIndex->hasDocument(m_docInfo.getLocation());
+	if (m_docId > 0)
+	{
+		// Ignore robots directives on updates
+		m_update = true;
+	}
+
 	// We may not have to download the document
 	// If coming from a crawl, this will be empty
 	if (m_docInfo.getType().empty() == true)
 	{
-		m_docInfo.setType(MIMEScanner::scanFile(m_docInfo.getLocation()));
+		m_docInfo.setType(MIMEScanner::scanUrl(thisUrl));
 	}
 
-	if (TokenizerFactory::isSupportedType(m_docInfo.getType(), dataNeeds) == false)
+	if (Dijon::FilterFactory::isSupportedType(m_docInfo.getType()) == false)
 	{
 		// Skip unsupported types ?
 		if (m_allowAllMIMETypes == false)
@@ -1168,27 +1331,35 @@ void IndexingThread::doWork(void)
 			return;
 		}
 	}
+	else
+	{
+		Dijon::Filter *pFilter = Dijon::FilterFactory::getFilter(m_docInfo.getType());
 
-	if ((dataNeeds == Tokenizer::ALL_BUT_FILES) &&
-		(thisUrl.getProtocol() == "file"))
-	{
-		doDownload = false;
-	}
-	else if (dataNeeds == Tokenizer::NO_DOCUMENTS)
-	{
-		doDownload = false;
+		if (pFilter != NULL)
+		{
+			// We may able to feed the document directly to the filter
+			if (((pFilter->is_data_input_ok(Dijon::Filter::DOCUMENT_FILE_NAME) == true) &&
+				(thisUrl.getProtocol() == "file")) ||
+				((pFilter->is_data_input_ok(Dijon::Filter::DOCUMENT_URI) == true) &&
+				(thisUrl.getProtocol() == "http")))
+			{
+				doDownload = false;
+			}
+
+			delete pFilter;
+		}
 	}
 
 	if (doDownload == true)
 	{
 		DownloadingThread::doWork();
-#ifdef DEBUG
-		cout << "IndexingThread::doWork: downloaded " << m_docInfo.getLocation() << endl;
-#endif
 	}
 	else
 	{
 		m_pDoc = new Document(m_docInfo);
+
+		m_pDoc->setTimestamp(m_docInfo.getTimestamp());
+		m_pDoc->setSize(m_docInfo.getSize());
 #ifdef DEBUG
 		cout << "IndexingThread::doWork: skipped download of " << m_docInfo.getLocation() << endl;
 #endif
@@ -1196,7 +1367,6 @@ void IndexingThread::doWork(void)
 
 	if (m_pDoc != NULL)
 	{
-		Tokenizer *pTokens = NULL;
 		string docType(m_pDoc->getType());
 		bool success = false;
 
@@ -1225,7 +1395,7 @@ void IndexingThread::doWork(void)
 #endif
 
 		// Check again as the downloader may have altered the MIME type
-		if (TokenizerFactory::isSupportedType(m_docInfo.getType(), dataNeeds) == false)
+		if (Dijon::FilterFactory::isSupportedType(m_docInfo.getType()) == false)
 		{
 			// Skip unsupported types ?
 			if (m_allowAllMIMETypes == false)
@@ -1239,59 +1409,57 @@ void IndexingThread::doWork(void)
 				m_status += m_docInfo.getLocation();
 				return;
 			}
-#ifdef DEBUG
-			cout << "IndexingThread::doWork: can't index document content" << endl;
-#endif
 
-			// Create an empty document so that the file's details are indexed but not its content
-			delete m_pDoc;
-			m_pDoc = new Document(m_docInfo);
-			// A simple text tokenizer will do the job
-			pTokens = new Tokenizer(m_pDoc);
+			// Let FilterWrapper handle unspported documents
 		}
-		else
+		else if ((PinotSettings::getInstance().m_ignoreRobotsDirectives == false) &&
+			(m_docInfo.getType().length() >= 9) &&
+			(m_docInfo.getType().substr(9) == "text/html"))
 		{
-			pTokens = TokenizerFactory::getTokenizerByType(m_docInfo.getType(), m_pDoc);
-		}
+			Dijon::HtmlFilter htmlFilter(m_docInfo.getType());
 
-		if (pTokens == NULL)
-                {
-                        m_status = _("Couldn't tokenize");
-                        m_status += " ";
-                        m_status += m_docInfo.getLocation();
-                        return;
-                }
-
-		// Is indexing allowed ?
-		HtmlTokenizer *pHtmlTokens = dynamic_cast<HtmlTokenizer*>(pTokens);
-		if ((m_ignoreRobotsDirectives == false) &&
-			(pHtmlTokens != NULL))
-		{
-			// See if the document has a ROBOTS META tag
-			string robotsDirectives = pHtmlTokens->getMetaTag("robots");
-			string::size_type pos1 = robotsDirectives.find("none");
-			string::size_type pos2 = robotsDirectives.find("noindex");
-			if ((pos1 != string::npos) ||
-				(pos2 != string::npos))
+			if ((FilterUtils::feedFilter(*m_pDoc, &htmlFilter) == true) &&
+				(htmlFilter.next_document() == true))
 			{
-				// No, it's not
-				delete pTokens;
-				m_status = _("Robots META tag forbids indexing");
-				m_status += " ";
-				m_status += m_docInfo.getLocation();
-				return;
+				const map<string, string> &metaData = htmlFilter.get_meta_data();
+
+				// See if the document has a ROBOTS META tag
+				map<string, string>::const_iterator robotsIter = metaData.find("robots");
+				if (robotsIter != metaData.end())
+				{
+					string robotsDirectives(robotsIter->second);
+	
+					// Is indexing allowed ?
+					string::size_type pos1 = robotsDirectives.find("none");
+					string::size_type pos2 = robotsDirectives.find("noindex");
+					if ((pos1 != string::npos) ||
+						(pos2 != string::npos))
+					{
+						// No, it isn't
+						m_status = _("Robots META tag forbids indexing");
+						m_status += " ";
+						m_status += m_docInfo.getLocation();
+						return;
+					}
+				}
 			}
+#ifdef DEBUG
+			else cout << "IndexingThread::doWork: couldn't check document for ROBOTS directive" << endl;
+#endif
 		}
 
 		if (m_done == false)
 		{
+			FilterWrapper wrapFilter(pIndex);
+			const set<string> &labels = m_docInfo.getLabels();
+
 			pIndex->setStemmingMode(IndexInterface::STORE_BOTH);
 
 			// Update an existing document or add to the index ?
 			if (m_update == true)
 			{
 				// Update the document
-				if (pIndex->updateDocument(m_docId, *pTokens) == true)
+				if (wrapFilter.updateDocument(*m_pDoc, m_docId) == true)
 				{
 #ifdef DEBUG
 					cout << "IndexingThread::doWork: updated " << m_pDoc->getLocation()
@@ -1305,11 +1473,10 @@ void IndexingThread::doWork(void)
 			}
 			else
 			{
-				const set<string> &labels = m_docInfo.getLabels();
 				unsigned int docId = 0;
 
 				// Index the document
-				success = pIndex->indexDocument(*pTokens, labels, docId);
+				success = wrapFilter.indexDocument(*m_pDoc, labels, docId);
 				if (success == true)
 				{
 					m_docId = docId;
@@ -1341,9 +1508,10 @@ void IndexingThread::doWork(void)
 				pIndex->getDocumentInfo(m_docId, m_docInfo);
 			}
 		}
-
-		delete pTokens;
 	}
+#ifdef DEBUG
+	else cout << "IndexingThread::doWork: couldn't download " << m_docInfo.getLocation() << endl;
+#endif
 
 	delete pIndex;
 }
@@ -1419,7 +1587,7 @@ void UnindexingThread::doWork(void)
 
 			// By unindexing all documents that match the label,
 			// we effectively delete the label from the index
-			if (pIndex->unindexDocuments(labelName) == true)
+			if (pIndex->unindexDocuments(labelName, IndexInterface::BY_LABEL) == true)
 			{
 #ifdef DEBUG
 				cout << "UnindexingThread::doWork: removed label " << labelName << endl;
@@ -1558,3 +1726,40 @@ void UpdateDocumentThread::doWork(void)
 	}
 }
 
+StartDaemonThread::StartDaemonThread() :
+	WorkerThread()
+{
+}
+
+StartDaemonThread::~StartDaemonThread()
+{
+}
+
+string StartDaemonThread::getType(void) const
+{
+	return "StartDaemonThread";
+}
+
+bool StartDaemonThread::stop(void)
+{
+	m_done = true;
+	return true;
+}
+
+void StartDaemonThread::doWork(void)
+{
+	unsigned int crawledCount = 0, docsCount = 0;
+
+	if (m_done == false)
+        {
+		// ... and let D-Bus activate the service if necessary
+		// If it was already running, changes will take effect when it's restarted
+		DBusXapianIndex::getStatistics(crawledCount, docsCount);
+
+#ifdef DEBUG
+		cout << "StartDaemonThread::doWork: crawled " << crawledCount
+			<< ", indexed " << docsCount << endl;
+#endif
+	}
+}
+
