@@ -16,6 +16,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+#include <stdio.h>
 #include <iostream>
 #include <utility>
 
@@ -31,18 +32,26 @@ using namespace Glib;
 using namespace Gtk;
 
 propertiesDialog::propertiesDialog(const std::set<std::string> &docLabels,
-	const DocumentInfo &docInfo, bool editDocument) :
+	const DocumentInfo &docInfo, unsigned int termsCount, bool editDocument) :
 	propertiesDialog_glade(),
 	m_editDocument(editDocument),
 	m_docInfo(docInfo)
 {
 	string language(docInfo.getLanguage());
+	char numStr[128];
+	unsigned int indexId = 0;
 	bool notALanguageName = false;
 
-	// Associate the columns model to the language combo
-	m_refLanguageTree = ListStore::create(m_languageColumns);
-	languageCombobox->set_model(m_refLanguageTree);
-	languageCombobox->pack_start(m_languageColumns.m_name);
+	// Get the document ID
+	unsigned int docId = docInfo.getIsIndexed(indexId);
+#ifdef DEBUG
+	cout << "propertiesDialog: document " << docId << endl;
+#endif
+	if (docId > 0)
+	{
+		snprintf(numStr, 128, "%u", docId);
+		set_title(get_title() + " (ID " + numStr + ")");
+	}
 
 	// Associate the columns model to the labels tree
 	m_refLabelsTree = ListStore::create(m_labelsColumns);
@@ -54,8 +63,28 @@ propertiesDialog::propertiesDialog(const std::set<std::string> &docLabels,
 
 	if (m_editDocument == true)
 	{
+		unsigned int size = docInfo.getSize();
+
 		titleEntry->set_text(to_utf8(docInfo.getTitle()));
 		typeEntry->set_text(to_utf8(docInfo.getType()));
+		if (size == 0)
+		{
+			sizeEntry->set_text(_("Unknown"));
+		}
+		else
+		{
+			snprintf(numStr, 128, "%u", size);
+			sizeEntry->set_text(numStr);
+		}
+		if (termsCount == 0)
+		{
+			termsEntry->set_text(_("Unknown"));
+		}
+		else
+		{
+			snprintf(numStr, 128, "%u", termsCount);
+			termsEntry->set_text(numStr);
+		}
 	}
 	else
 	{
@@ -69,6 +98,10 @@ propertiesDialog::propertiesDialog(const std::set<std::string> &docLabels,
 		titleEntry->hide();
 		typeLabel->hide();
 		typeEntry->hide();
+		sizeLabel->hide();
+		sizeEntry->hide();
+		termsLabel->hide();
+		termsEntry->hide();
 	}
 
 	populate_languageCombobox(language, notALanguageName);
@@ -81,18 +114,12 @@ propertiesDialog::~propertiesDialog()
 
 void propertiesDialog::populate_languageCombobox(const string &language, bool notALanguageName)
 {
-	TreeModel::iterator iter;
-	TreeModel::Row row;
 	unsigned int languageStart = 0;
 	bool foundLanguage = false;
 
 	if (notALanguageName == true)
 	{
-		iter = m_refLanguageTree->append();
-		row = *iter;
-
-		// This is not a language name as such
-		row[m_languageColumns.m_name] = language;
+		languageCombobox->append_text(to_utf8(language));
 		languageCombobox->set_active(0);
 		languageStart = 1;
 	}
@@ -100,11 +127,9 @@ void propertiesDialog::populate_languageCombobox(const string &language, bool no
 	// Add all supported languages
 	for (unsigned int languageNum = 0; languageNum < Languages::m_count; ++languageNum)
 	{
-		string languageName = Languages::getIntlName(languageNum);
-		iter = m_refLanguageTree->append();
-		row = *iter;
-		row[m_languageColumns.m_name] = languageName;
+		string languageName(Languages::getIntlName(languageNum));
 
+		languageCombobox->append_text(to_utf8(languageName));
 		if ((notALanguageName == false) &&
 			(languageName == language))
 		{
@@ -165,7 +190,7 @@ void propertiesDialog::setHeight(int maxHeight)
 		// What's the current size ?
 		get_size(width, height);
 		// Add enough room for the rows we need to show
-		height += get_column_height(labelsTreeview) * (rowsCount - 2);
+		height += get_column_height(labelsTreeview) * (rowsCount - 1);
 		// Resize
 		resize(width, min(maxHeight, height));
 	}
@@ -198,11 +223,7 @@ void propertiesDialog::on_labelOkButton_clicked()
 	int chosenLanguage = languageCombobox->get_active_row_number();
 	if (chosenLanguage >= languageStart)
 	{
-		TreeModel::iterator iter = languageCombobox->get_active();
-		TreeModel::Row row = *iter;
-		string languageName = from_utf8(row[m_languageColumns.m_name]);
-
-		m_docInfo.setLanguage(languageName);
+		m_docInfo.setLanguage(from_utf8(languageCombobox->get_active_text()));
 	}
 
 	// Go through the labels tree
